"""
Wrapper for interactions with shared database that holds arbitrary data that persists across the entire recipe run
"""
import json
from collections.abc import MutableMapping
from typing import Generator
from typing import Union

from dkist_processing_common._util.tags import TagDB


class Constants(MutableMapping):
    def __init__(self, recipe_run_id: int, task_name: str):
        """
        Initialize a connection to the shared database
        """
        self.store = TagDB(recipe_run_id, task_name, "constant")

    @staticmethod
    def extract_value(value: set) -> Union[int, str, float]:
        """
        Pop the first (and only) value from set and convert it from a json string
        """
        return json.loads(value.pop())

    def __getitem__(self, key: str) -> Union[int, str, float]:
        """
        Return the constant stored at a specific key. Raise and error if the key doesn't exist.
        """
        value = self.store.all(key)
        if not value:
            raise KeyError(f"Constant {key} does not exist")
        return self.extract_value(value)

    def __delitem__(self, key: str):
        """
        "Delete" a key by making it map to an empty set
        """
        self.store.clear_tag(key)

    def __setitem__(self, key: str, value: Union[str, int, float]):
        """
        Set a constant key with the specified value. Raise an error if the key already exists.
        """
        if self.store.all(key):
            raise ValueError(f"Constant {key} already exists")
        self.store.add(key, json.dumps(value))

    def __iter__(self) -> Generator[str, None, None]:
        """
        Yield the currently defined constants as strings
        """
        yield from self.store.tags

    def __len__(self):
        """
        The number of constants currently defined
        """
        return len(self.store.tags)

    def close(self):
        """
        Close the db connection.  Call on __exit__ of a Task
        """
        self.store.close()

    def purge(self):
        """
        Remove all constants associated with the instance recipe run id
        """
        self.store.purge()



"""
Wrappers for all workflow tasks
"""
import json
from abc import ABC
from io import BytesIO
from pathlib import Path
from string import ascii_uppercase
from typing import Generator
from typing import Iterable
from typing import List
from typing import Union
from uuid import uuid4

import pkg_resources
from dkist_processing_core import TaskBase
from hashids import Hashids

from dkist_processing_common._util.constants import Constants
from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common.models.constants import BudName
from dkist_processing_common.tasks.mixin.fits import FitsDataMixin
from dkist_processing_common.tasks.mixin.metadata_store import MetadataStoreMixin

__all__ = ["ParsedL0InputTaskBase", "ScienceTaskL0ToL1Base", "WorkflowDataTaskBase"]

tag_type_hint = Union[Iterable[str], str]


class WorkflowDataTaskBase(TaskBase, ABC):
    """
    Wrapper for all tasks that need to access the persistent automated processing data stores.
    Adds capabilities for accessing:
      scratch
      tags
      constants
    """

    def __init__(
        self,
        recipe_run_id: int,
        workflow_name: str,
        workflow_version: str,
    ):
        super().__init__(
            recipe_run_id=recipe_run_id,
            workflow_name=workflow_name,
            workflow_version=workflow_version,
        )
        task_name = self.__class__.__name__
        self.scratch = WorkflowFileSystem(recipe_run_id=recipe_run_id, task_name=task_name)
        self.constants = Constants(recipe_run_id=recipe_run_id, task_name=task_name)

    def read(self, tags: tag_type_hint) -> Generator[Path, None, None]:
        tags = self._parse_tags(tags)
        return self.scratch.find_all(tags=tags)

    def write(
        self,
        file_obj: Union[BytesIO, bytes],
        tags: tag_type_hint,
        relative_path: Union[Path, str, None] = None,
    ) -> Path:
        if not tags:
            raise ValueError(f"Tags are required")
        if isinstance(file_obj, BytesIO):
            file_obj = file_obj.read()
        tags = self._parse_tags(tags)
        relative_path = relative_path or f"{uuid4().hex}.dat"
        relative_path = Path(relative_path)
        self.scratch.write(file_obj=file_obj, relative_path=relative_path, tags=tags)
        return relative_path

    def tag(self, path: Union[Path, str], tags: tag_type_hint) -> None:
        """
        Wrapper for the tag method in WorkflowFileSystem
        """
        tags = self._parse_tags(tags)
        return self.scratch.tag(path=path, tags=tags)

    def tags(self, path: Union[Path, str]) -> List[str]:
        """
        Return list of tags that a path belongs to
        """
        return self.scratch.tags(path=path)

    @staticmethod
    def _parse_tags(tags: tag_type_hint) -> Iterable[str]:
        result = []
        if isinstance(tags, str):
            tags = [tags]
        for tag in tags:
            if not isinstance(tag, str):
                raise TypeError(f"Tags must be strings. Got {type(tag)} instead.")
            result.append(tag)
        return result

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        self.scratch.close()
        self.constants.close()


class ParsedL0InputTaskBase(WorkflowDataTaskBase, ABC):
    @property
    def proposal_id(self) -> str:
        return self.constants[BudName.proposal_id.value]

    @property
    def dataset_id(self) -> str:
        return Hashids(min_length=5, alphabet=ascii_uppercase).encode(self.recipe_run_id)

    @property
    def instrument(self) -> str:
        return self.constants[BudName.instrument.value]

    @property
    def average_cadence(self) -> float:
        return self.constants[BudName.average_cadence.value]

    @property
    def maximum_cadence(self) -> float:
        return self.constants[BudName.maximum_cadence.value]

    @property
    def minimum_cadence(self) -> float:
        return self.constants[BudName.minimum_cadence.value]

    @property
    def variance_cadence(self) -> float:
        return self.constants[BudName.variance_cadence.value]

    @property
    def time_order(self) -> [float]:
        return json.loads(self.constants[BudName.time_order.value])

    @property
    def num_dsps_repeats(self) -> int:
        return self.constants[BudName.num_dsps_repeats.value]

    @property
    def spectral_line(self) -> str:
        return self.constants[BudName.spectral_line.value]


class ScienceTaskL0ToL1Base(ParsedL0InputTaskBase, MetadataStoreMixin, FitsDataMixin, ABC):
    """"""

    is_task_manual: bool = False

    @property
    def library_versions(self) -> str:
        """
        Harvest the dependency names and versions from the environment for
          all packages beginning with 'dkist' or are a requirement for a package
          beginning with 'dkist'
        """
        distributions = {d.key: d.version for d in pkg_resources.working_set}
        libraries = {}
        for pkg in pkg_resources.working_set:
            if pkg.key.startswith("dkist"):
                libraries[pkg.key] = pkg.version
                for req in pkg.requires():
                    libraries[req.key] = distributions[req.key]
        return json.dumps(libraries)

    def record_provenance(self):
        self.metadata_store_record_provenance(
            is_task_manual=self.is_task_manual, library_versions=self.library_versions
        )

    def pre_run(self) -> None:
        super().pre_run()
        with self.apm_step("Record Provenance"):
            self.record_provenance()



from dkist_processing_common.models.constants import BudName
from dkist_processing_common.models.flower_pot import SpilledDirt
from dkist_processing_common.models.tags import StemName
from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess
from dkist_processing_common.parsers.single_value_single_key_flower import (
    SingleValueSingleKeyFlower,
)
from dkist_processing_common.parsers.unique_bud import UniqueBud


class TotalDspsRepeatsBud(UniqueBud):
    def __init__(self):
        super().__init__(
            constant_name=BudName.num_dsps_repeats.value, metadata_key="num_dsps_repeats"
        )

    def setter(self, fits_obj: L0FitsAccess):
        if fits_obj.ip_task_type != "observe":
            return SpilledDirt
        return super().setter(fits_obj)


class DspsRepeatNumberFlower(SingleValueSingleKeyFlower):
    def __init__(self):
        super().__init__(
            tag_stem_name=StemName.dsps_repeat.value, metadata_key="current_dsps_repeat"
        )

    def setter(self, fits_obj: L0FitsAccess):
        if fits_obj.ip_task_type != "observe":
            return SpilledDirt
        return super().setter(fits_obj)



"""
Pre-made flower that reads a single header key from all files and raises a ValueError if it is not unique
"""
from dkist_processing_common.models.flower_pot import Stem
from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess


class UniqueBud(Stem):
    def __init__(
        self,
        constant_name: str,
        metadata_key: str,
    ):
        super().__init__(stem_name=constant_name)
        self.metadata_key = metadata_key

    def setter(self, fits_obj: L0FitsAccess):
        return getattr(fits_obj, self.metadata_key)

    def getter(self, key):
        value_set = set(self.key_to_petal_dict.values())
        if len(value_set) > 1:
            raise ValueError(
                f"Multiple {self.stem_name} values found for key {key}. Values: {value_set}"
            )
        return value_set.pop()



"""
Tests for the tag cloud manager
"""
import random
from typing import Callable

import pytest


@pytest.fixture()
def tag_data(tag_db):
    tags = ["A", "B", "C"]
    intersection = "Intersect"
    union = {f"Union{t}" for t in tags}
    union.update({intersection})
    for t in tags:
        tag_db.add(t, f"Union{t}")
        tag_db.add(t, intersection)
    return tags, {intersection}, union


@pytest.fixture()
def tag_db2_data(tag_db2):
    tags = ["D", "E", "F"]
    for t in tags:
        tag_db2.add(t, f"value_of_{t}")
    return tags


def test_tag_db_add(tag_db):
    """
    Given: Connection to a tag database
    When: Adding a tag
    Then: Tag is retrievable
    """
    tag = "ADD"
    values = [f"add{x}" for x in range(100)]
    # When
    for v in values:
        tag_db.add(tag, v)
    # Then
    assert tag_db.any(tag) == set(values)


def test_tag_db_clear_tag(tag_db, tag_data):
    """
    Given: Connection to a tag database with data
    When: Clearing a tag
    Then: Cleared tag returns an empty set, but other keys are not affected
    """
    tags, intersection, _ = tag_data
    tag_idx_to_clear = random.randint(0, len(tags) - 1)
    tag_to_clear = tags[tag_idx_to_clear]
    del tags[tag_idx_to_clear]
    union = {f"Union{t}" for t in tags}
    union.update(intersection)
    tag_db.clear_tag(tag_to_clear)
    assert len(tag_db.any(tag_to_clear)) == 0
    assert tag_db.any(tags) == union
    assert tag_db.all(tags) == intersection


def test_tag_db_any(tag_db, tag_data):
    """
    Given: Connection to a tag database with data
    When: When searching for the results of any tags
    Then: Union of values returned
    """
    tags, _, union = tag_data
    # When
    result = tag_db.any(tags)
    # Then
    assert result == union


def test_tag_db_all(tag_db, tag_data):
    """
    Given: Connection to a tag database with data
    When: When searching for the results of all tags
    Then: Intersection of values returned
    """
    tags, intersection, _ = tag_data
    # When
    result = tag_db.all(tags)
    # Then
    assert result == intersection


@pytest.mark.parametrize(
    "func, attr",
    [
        pytest.param(repr, "__repr__", id="repr"),
        pytest.param(str, "__str__", id="str"),
    ],
)
def test_tag_db_dunder(tag_db, func: Callable, attr):
    """
    Given: Connection to a tag database
    When: retrieving dunder method that should be implemented
    Then: It is implemented
    """
    assert getattr(tag_db, attr, None)
    assert func(tag_db)


def test_tag_db_purge(tag_db, tag_db2, tag_data, tag_db2_data):
    """
    Given: Given a connection to a 2 tag_db instances with data
        in the same redis db (have same db # but different namespace)
    When: tag_db 1 is purged
    Then: No tags remain for the tag_db 1 instance; tags still remain in tag_db 2
    """
    # When
    tag_db.purge()
    # Then
    assert not tag_db._namespace_keys
    assert len(tag_db2._namespace_keys) == len(tag_db2_data)


def test_tag_db_tags(tag_db):
    """
    Given: A connection to a tag db with a single tag
    When: Asking for a list of tags that are defined
    Then: The correct values are returned as a list of strings
    """
    tag_db.add("TAG", "value")
    tags = tag_db.tags
    assert tags == ["TAG"]


def test_tag_db_clear_value(tag_db, tag_data):
    """
    Given: A connection to a tag db with multiple tags that each have values
    When: clearing a known value from all tags
    Then: that value is not found in any of the tags
    """
    tags, _, _ = tag_data
    tag_db.clear_value(value="Intersect")
    assert "Intersect" not in tag_db.any(tags=tags)


def test_tags_for_value(tag_db, tag_data):
    """
    Given: A connection to a tag db with multiple tags that each have values
    When: asking for the tags associated with each value
    Then: the correct tag list is returned for each value
    """
    tags, intersection, union = tag_data
    intersect_val = intersection.pop()
    assert sorted(tag_db.tags_for_value(intersect_val)) == tags
    for v in union:
        if v == intersect_val:
            continue
        inferred_tag = v[-1]
        assert tag_db.tags_for_value(v) == [inferred_tag]



from __future__ import annotations

from typing import Optional
from typing import Type
from typing import Union

from astropy.time import Time

from dkist_processing_common.models.constants import BudName
from dkist_processing_common.models.flower_pot import SpilledDirt
from dkist_processing_common.models.flower_pot import Stem
from dkist_processing_common.models.tags import StemName
from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess


class CSStep:
    """This class allows for an easy way to quickly compare fits access objects to determine if they come from the
    same Calibration Sequence (CS) step. Each step in a CS is defined by the configuration of the GOS, namely the
    status of the polarizer and retarder (in or out) and the angle of each. Because some CS schemes call for some
    GOS configurations to repeat a check is also made against the observation time for each object; the default in
    the parameter store defines a maximum difference in time where two exposures are considered different CS steps
    regardless of GOS configuration.

    This class can also be sorted. In this case, only the observation time is taken into account.

    Finally, this class is hashable for use in dictionaries. The hash is only based on the GOS configuration so that
    all objects from the same CS step result in the same hash. Python also checks that dictionary keys are equal so any
    conflicts with repeated CS steps are avoided.
    """

    def __init__(self, fits_obj: L0FitsAccess, max_cs_time_sec: Optional[float] = None):
        """Initialize and read the GOS configuration and time from a SPEC-0122 FITS header.

        Parameters
        ----------
        fits_obj
            The FitsAccess object to ingest
        max_cs_time_sec
            Any objects taken more than this amount of time apart are considered in separate steps, regardless of GOS
            configuration.
        """
        self.pol_in = fits_obj.gos_polarizer_status not in ["undefined", "clear"]
        self.pol_theta = fits_obj.gos_polarizer_angle
        self.ret_in = fits_obj.gos_retarder_status not in ["undefined", "clear"]
        self.ret_theta = fits_obj.gos_retarder_angle
        self.dark_in = fits_obj.gos_level0_status == "DarkShutter"
        self.obs_time = Time(fits_obj.time_obs)
        self.max_cs_time_sec = max_cs_time_sec

        self.max_cs_time_sec = (
            max_cs_time_sec or 20.0
        )  # Default mostly (and hopefully only) here for testing

    def __repr__(self):
        return f"CS step taken on {self.obs_time.fits}"

    def __str__(self):
        return "CS step with Pol = {}:{}, Ret = {}:{}, Dark = {}. Taken at {}".format(
            self.pol_in,
            self.pol_theta,
            self.ret_in,
            self.ret_theta,
            self.dark_in,
            self.obs_time.fits,
        )

    def __eq__(self, other: object) -> bool:
        """Two steps are equal if they have the same GOS configuration and are taken within some package-defined time
        of each other."""
        if not isinstance(other, CSStep):
            raise TypeError(f"Cannot compare CSStep with type {type(other)}")

        for item in ["pol_in", "pol_theta", "ret_in", "ret_theta", "dark_in"]:
            if getattr(self, item) != getattr(other, item):
                return False

        tdelt = abs(self.obs_time - other.obs_time)
        if tdelt.sec > self.max_cs_time_sec:
            return False

        return True

    def __lt__(self, other: CSStep) -> bool:
        """Only based on time"""
        return self.obs_time < other.obs_time

    def __hash__(self) -> int:
        """Only based on GSO configuration so that all objects from the same CS step hash the same"""
        return hash((self.pol_in, self.pol_theta, self.ret_in, self.ret_theta, self.dark_in))


class CSStepFlower(Stem):
    """ Identify which CS Step a header belongs to """

    def __init__(self, max_cs_step_time_sec: Optional[float] = None):
        super().__init__(stem_name=StemName.cs_step.value)
        self.max_cs_step_time_sec = max_cs_step_time_sec

    def setter(self, fits_obj: L0FitsAccess) -> Union[CSStep, Type[SpilledDirt]]:
        if fits_obj.ip_task_type != "polcal":
            return SpilledDirt
        return CSStep(fits_obj, max_cs_time_sec=self.max_cs_step_time_sec)

    def getter(self, key) -> Union[str, float, int]:
        unique_steps = sorted(list(set(self.key_to_petal_dict.values())))
        return unique_steps.index(self.key_to_petal_dict[key])


class NumCSStepBud(Stem):
    """ The total number of CS Steps present in a dataset """

    def __init__(self, max_cs_step_time_sec: Optional[float] = None):
        super().__init__(stem_name=BudName.num_cs_steps.value)
        self.max_cs_step_time_sec = max_cs_step_time_sec

    def setter(self, fits_obj: L0FitsAccess) -> Union[CSStep, Type[SpilledDirt]]:
        if fits_obj.ip_task_type != "polcal":
            return SpilledDirt
        return CSStep(fits_obj, max_cs_time_sec=self.max_cs_step_time_sec)

    def getter(self, key) -> int:
        value_set = set(self.key_to_petal_dict.values())
        return len(value_set)



import json
from typing import Union

import numpy as np
from astropy.time import Time

from dkist_processing_common.models.constants import BudName
from dkist_processing_common.models.flower_pot import SpilledDirt
from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess
from dkist_processing_common.parsers.unique_bud import UniqueBud


class TimeBud(UniqueBud):
    """
    Base class for all Time Buds
    """

    def __init__(self, constant_name: str):
        super().__init__(constant_name, metadata_key="time_obs")

    def setter(self, fits_obj: L0FitsAccess) -> Union[float, SpilledDirt]:
        """
        If the file is an observe file, its DATE-OBS value is stored as unix seconds
        """
        if fits_obj.ip_task_type == "observe":
            return Time(getattr(fits_obj, self.metadata_key)).unix
        return SpilledDirt


class AverageCadenceBud(TimeBud):
    def __init__(self):
        super().__init__(constant_name=BudName.average_cadence.value)

    def getter(self, key) -> np.float64:
        """
        Return the mean cadence between frames
        """
        return np.mean(np.diff(sorted(list(self.key_to_petal_dict.values()))))


class MaximumCadenceBud(TimeBud):
    def __init__(self):
        super().__init__(constant_name=BudName.maximum_cadence.value)

    def getter(self, key) -> np.float64:
        """
        Return the maximum cadence between frames
        """
        return np.max(np.diff(sorted(list(self.key_to_petal_dict.values()))))


class MinimumCadenceBud(TimeBud):
    def __init__(self):
        super().__init__(constant_name=BudName.minimum_cadence.value)

    def getter(self, key) -> np.float64:
        """
        Return the minimum cadence between frames
        """
        return np.min(np.diff(sorted(list(self.key_to_petal_dict.values()))))


class VarianceCadenceBud(TimeBud):
    def __init__(self):
        super().__init__(constant_name=BudName.variance_cadence.value)

    def getter(self, key) -> np.float64:
        """
        Return the cadence variance between frames
        """
        return np.var(np.diff(sorted(list(self.key_to_petal_dict.values()))))


class TimeOrderBud(TimeBud):
    def __init__(self):
        super().__init__(constant_name=BudName.time_order.value)

    def getter(self, key) -> str:
        """
        Return the times as a json string that can be decoded to a list of floats
        """
        return json.dumps(sorted(list(self.key_to_petal_dict.values())))



import json
import logging
from configparser import ConfigParser
from pathlib import Path

import pytest

import dkist_processing_common
from dkist_processing_common.tasks.base import ScienceTaskL0ToL1Base


class Task(ScienceTaskL0ToL1Base):
    def run(self):
        ...


@pytest.fixture(scope="function")
def science_l0_task(tmp_path, recipe_run_id):
    with Task(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        yield task


@pytest.fixture()
def package_dependencies() -> set:
    """
    Extract dependencies from setup.cfg and format into a set of package names
    """
    module_path = Path(dkist_processing_common.__path__[0])
    setup_cfg = module_path.parent / "setup.cfg"
    logging.info(setup_cfg)
    config = ConfigParser()
    config.read(setup_cfg)
    install_requires = [d for d in config["options"]["install_requires"].splitlines() if d]
    requirements = install_requires + ["dkist-processing-common"]
    dependencies = {pkg.split(" ")[0] for pkg in requirements}
    dependencies_without_optionals = {d.split("[")[0] for d in dependencies}
    return dependencies_without_optionals


def test_library_versions(science_l0_task, package_dependencies):
    """
    Given: An instance of a TaskBase subclass
    When: accessing library_versions attr
    Then: Result contains the dkist-processing-core package and packages specified in setup.cfg
      - options.install_requires
      - options.extras_require.test
      Result does not contain any other packages
      Result structure is Dict[str,str] where the key is library name and value is the version
    """
    library_names = json.loads(science_l0_task.library_versions).keys()
    assert len(library_names) == len(set(library_names))  # no duplicates
    for package in package_dependencies:
        assert package in set(library_names)
    library_versions = json.loads(science_l0_task.library_versions).values()
    assert all(library_versions)  # values for all versions



from typing import Union

from dkist_processing_common.models.constants import BudName
from dkist_processing_common.models.flower_pot import SpilledDirt
from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess
from dkist_processing_common.parsers.unique_bud import UniqueBud


class ProposalIdBud(UniqueBud):
    def __init__(self):
        super().__init__(constant_name=BudName.proposal_id.value, metadata_key="proposal_id")

    def setter(self, fits_obj: L0FitsAccess) -> Union[str, SpilledDirt]:
        if fits_obj.ip_task_type == "observe":
            return getattr(fits_obj, self.metadata_key)
        return SpilledDirt



from string import ascii_uppercase

import pytest
from hashids import Hashids
from hypothesis import given
from hypothesis.strategies import integers

from dkist_processing_common.tasks.base import ParsedL0InputTaskBase


class Task(ParsedL0InputTaskBase):
    def run(self):
        pass


@pytest.fixture
def parsed_l0(recipe_run_id):
    proposal_id = "proposal_id"
    with Task(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        task.constants["PROPOSAL_ID"] = proposal_id
        yield task, proposal_id
        task.constants.purge()


def test_proposal_id(parsed_l0):
    """
    Given: a ParsedL0InputTaskBase task
    When: getting the proposal id
    Then: the proposal id from the constants mutable mapping is returned
    """
    task, proposal_id = parsed_l0
    task()
    assert task.proposal_id == proposal_id


def test_dataset_id(parsed_l0):
    """
    Given: a ParsedL0InputTaskBase task
    When: getting the dataset id
    Then: the dataset id hashed from the recipe run id is returned
    """
    task, _ = parsed_l0
    assert task.dataset_id == Hashids(min_length=5, alphabet=ascii_uppercase).encode(
        task.recipe_run_id
    )


@given(id_x=integers(min_value=1), id_y=integers(min_value=1))
def test_dataset_id_uniquely_generated_from_recipe_run_id(id_x, id_y):
    """
    Given: 2 integers > 0
    When: 2 tasks are created using the integers for each of the tasks
    Then: The dataset_id for each class compares the same as the integers compare
       e.g. (1 == 1) is (dataset_id(1) == dataset_id(1))  : True is True
       e.g. (1 == 2) == (dataset_id(1) == dataset_id(2))  : False is False
    """
    expected = id_x == id_y
    task_x = Task(recipe_run_id=id_x, workflow_name="", workflow_version="")
    task_y = Task(recipe_run_id=id_y, workflow_name="", workflow_version="")
    actual = task_x.dataset_id == task_y.dataset_id
    assert expected is actual


@given(id_x=integers(min_value=1))
def test_dataset_id_from_recipe_run_id_produces_the_same_value(id_x):
    """
    Given: an integer > 0
    When: 2 tasks are created using the same integer for each of the tasks
    Then: The dataset_id for each class are equal
    """
    task_1 = Task(recipe_run_id=id_x, workflow_name="", workflow_version="")
    task_2 = Task(recipe_run_id=id_x, workflow_name="", workflow_version="")
    assert task_1.dataset_id == task_2.dataset_id



from typing import Tuple

import numpy as np
import pytest
from astropy.io import fits
from dkist_data_simulator.dataset import key_function
from dkist_data_simulator.spec122 import Spec122Dataset

from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common.models.constants import BudName
from dkist_processing_common.models.fits_access import FitsAccessBase
from dkist_processing_common.models.flower_pot import SpilledDirt
from dkist_processing_common.models.flower_pot import Stem
from dkist_processing_common.models.tags import StemName
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.parsers.single_value_single_key_flower import (
    SingleValueSingleKeyFlower,
)
from dkist_processing_common.parsers.unique_bud import UniqueBud
from dkist_processing_common.tasks.parse_l0_input_data import ParseL0InputData


class VispHeaders(Spec122Dataset):
    def __init__(
        self,
        num_mod: int,
        num_files_per_mod: int,
        array_shape: Tuple[int, ...],
        time_delta: float,
        instrument="visp",
    ):
        self.num_frames = num_mod * num_files_per_mod
        super().__init__(
            (self.num_frames, *array_shape[1:]), array_shape, time_delta, instrument=instrument
        )
        self.num_mod = num_mod
        self.add_constant_key("WAVELNTH")
        self.add_constant_key("TELSCAN")
        self.add_constant_key("CAM__001")
        self.add_constant_key("CAM__002")
        self.add_constant_key("CAM__003")
        self.add_constant_key("CAM__004")
        self.add_constant_key("CAM__005")
        self.add_constant_key("CAM__006")
        self.add_constant_key("CAM__007")
        self.add_constant_key("CAM__008")
        self.add_constant_key("CAM__009")
        self.add_constant_key("CAM__010")
        self.add_constant_key("CAM__011")
        self.add_constant_key("CAM__012")
        self.add_constant_key("CAM__013")
        self.add_constant_key("CAM__014")
        self.add_constant_key("CAM__015")
        self.add_constant_key("CAM__016")
        self.add_constant_key("CAM__017")
        self.add_constant_key("CAM__018")
        self.add_constant_key("CAM__019")
        self.add_constant_key("CAM__020")
        self.add_constant_key("CAM__021")
        self.add_constant_key("CAM__022")
        self.add_constant_key("CAM__023")
        self.add_constant_key("CAM__024")
        self.add_constant_key("CAM__025")
        self.add_constant_key("CAM__026")
        self.add_constant_key("CAM__027")
        self.add_constant_key("CAM__028")
        self.add_constant_key("CAM__029")
        self.add_constant_key("CAM__030")
        self.add_constant_key("CAM__031")
        self.add_constant_key("CAM__032")
        self.add_constant_key("PAC__002")
        self.add_constant_key("PAC__004")
        self.add_constant_key("PAC__006")
        self.add_constant_key("PAC__008")
        self.add_constant_key("VISP_002")
        self.add_constant_key("VISP_007")
        self.add_constant_key("VISP_010", self.num_mod)
        self.add_constant_key("VISP_016")
        self.add_constant_key("VISP_019")

    @key_function("VISP_011")
    def modstate(self, key: str) -> str:
        return self.index % self.num_mod


class ViSPFitsAccess(FitsAccessBase):
    def __init__(self, hdu, name):
        super().__init__(hdu, name)
        self.num_mod: int = self.header["VISP_010"]
        self.modstate: int = self.header["VISP_011"]
        self.ip_task_type: str = self.header["DKIST004"]
        self.name = name


@pytest.fixture(scope="function")
def visp_flowers():
    return [
        SingleValueSingleKeyFlower(tag_stem_name=StemName.modstate.value, metadata_key="modstate")
    ]


@pytest.fixture(scope="function")
def visp_buds():
    return [UniqueBud(constant_name=BudName.num_modstates.value, metadata_key="num_mod")]


@pytest.fixture(scope="function")
def empty_flowers():
    class EmptyFlower(Stem):
        def __init__(self):
            super().__init__(stem_name="EMPTY_FLOWER")

        def setter(self, value):
            return SpilledDirt

        def getter(self, key):
            pass  # We'll never get here because we spilled the dirt

    return [EmptyFlower()]


@pytest.fixture(scope="function")
def empty_buds():
    class EmptyBud(Stem):
        def __init__(self):
            super().__init__(stem_name="EMPTY_BUD")

        def setter(self, value):
            return SpilledDirt

        def getter(self, key):
            pass  # We'll never get here because we spilled the dirt

    return [EmptyBud()]


@pytest.fixture(scope="function")
def parse_inputs_task(tmp_path, visp_flowers, visp_buds, empty_flowers, empty_buds):
    class TaskClass(ParseL0InputData):
        @property
        def tag_flowers(self):
            return visp_flowers + empty_flowers

        @property
        def constant_flowers(self):
            return visp_buds + empty_buds

        @property
        def fits_parsing_class(self):
            return ViSPFitsAccess

        def run(self):
            pass

    with TaskClass(
        recipe_run_id=1, workflow_name="parse_visp_input_data", workflow_version="VX.Y"
    ) as task:
        task.scratch = WorkflowFileSystem(scratch_base_path=tmp_path)
        task._num_mod = 2
        task._num_files_per_mod = 3
        ds = VispHeaders(
            num_mod=task._num_mod,
            num_files_per_mod=task._num_files_per_mod,
            array_shape=(1, 512, 512),
            time_delta=10,
        )
        header_generator = (d.header() for d in ds)
        for i in range(ds.num_frames):
            hdu = fits.PrimaryHDU(data=np.zeros(shape=(1, 10, 10)))
            generated_header = next(header_generator)
            for key, value in generated_header.items():
                hdu.header[key] = value
            hdul = fits.HDUList([hdu])
            task.fits_data_write(
                hdu_list=hdul,
                tags=[Tag.input(), Tag.frame()],
                relative_path=f"input/input_{i}.fits",
            )
        yield task
        task.scratch.purge()
        task.constants.purge()


@pytest.fixture()
def visp_parse_inputs_task(tmp_path, visp_flowers, visp_buds):
    class TaskClass(ParseL0InputData):
        @property
        def tag_flowers(self):
            return super().tag_flowers + visp_flowers

        @property
        def constant_flowers(self):
            return super().constant_flowers + visp_buds

        @property
        def fits_parsing_class(self):
            return ViSPFitsAccess

        def run(self):
            pass

    with TaskClass(
        recipe_run_id=1, workflow_name="parse_visp_input_data", workflow_version="VX.Y"
    ) as task:
        yield task
        task.scratch.purge()
        task.constants.purge()


def test_make_flowerpots(parse_inputs_task):
    """
    Given: ParseInputData task with constant and tag Flowers
    When: Constructing constant and tag FlowerPots
    Then: The Flowers associated with the Task are correctly placed in either FlowerPot
    """

    tag_pot, constant_pot = parse_inputs_task.make_flower_pots()

    assert len(tag_pot.flowers) == 2
    assert len(constant_pot.flowers) == 2
    assert tag_pot.flowers[0].stem_name == StemName.modstate.value
    assert tag_pot.flowers[1].stem_name == "EMPTY_FLOWER"
    assert constant_pot.flowers[0].stem_name == BudName.num_modstates.value
    assert constant_pot.flowers[1].stem_name == "EMPTY_BUD"


def test_subclass_flowers(visp_parse_inputs_task):
    """
    Given: ParseInputData child class with custom flowers
    When: Making the flower pots
    Then: Both the base and custom flowers are placed in the correct FlowerPots
    """
    tag_pot, constant_pot = visp_parse_inputs_task.make_flower_pots()

    assert len(tag_pot.flowers) == 2
    assert len(constant_pot.flowers) == 9
    assert sorted([f.stem_name for f in tag_pot.flowers]) == sorted(
        [StemName.cs_step.value, StemName.modstate.value]
    )
    assert sorted([f.stem_name for f in constant_pot.flowers]) == sorted(
        [
            BudName.instrument.value,
            BudName.num_modstates.value,
            BudName.num_cs_steps.value,
            BudName.proposal_id.value,
            BudName.average_cadence.value,
            BudName.maximum_cadence.value,
            BudName.minimum_cadence.value,
            BudName.variance_cadence.value,
            BudName.time_order.value,
        ]
    )


def test_constants_correct(parse_inputs_task):
    """
    Given: ParseInputData task with a populated constant FlowerPot
    When: Updating pipeline constants
    Then: Pipeline constants are correctly populated
    """
    _, constant_pot = parse_inputs_task.make_flower_pots()
    parse_inputs_task.update_constants(constant_pot)
    assert parse_inputs_task.constants == {BudName.num_modstates.value: parse_inputs_task._num_mod}


def test_tags_correct(parse_inputs_task):
    """
    Given: ParseInputData task with a populated tag FlowerPot
    When: Tagging files with group information
    Then: All files are correctly tagged
    """
    tag_pot, _ = parse_inputs_task.make_flower_pots()
    parse_inputs_task.tag_petals(tag_pot)
    num_mod = parse_inputs_task._num_mod
    files_per_mod = parse_inputs_task._num_files_per_mod
    expected_tag_set = {Tag.input(), Tag.frame()}
    for m in range(num_mod):
        expected_tag_set.add(Tag.modstate(m))
        expected_mod_files = [
            parse_inputs_task.scratch.absolute_path(f"input/input_{i}.fits")
            for i in range(num_mod * files_per_mod)[m::num_mod]
        ]
        assert sorted(list(parse_inputs_task.read(tags=Tag.modstate(m)))) == expected_mod_files

    # To make sure the empty flower didn't make it in to the tags
    assert set(parse_inputs_task.scratch._tag_db.tags) == expected_tag_set



"""
Abstraction layer for accessing fits data via class attributes
"""
from __future__ import annotations

from typing import Optional
from typing import Union

import numpy as np
from astropy.io import fits


class FitsAccessBase:
    def __init__(
        self,
        hdu: Union[fits.ImageHDU, fits.PrimaryHDU, fits.CompImageHDU],
        name: Optional[str] = None,
        auto_squeeze: bool = True,
    ):
        self._hdu = hdu
        self.name = name
        self.auto_squeeze = auto_squeeze

    @property
    def data(self) -> np.ndarray:
        """
        Return the data array from the associated FITS object, with axes of length 1 removed if the
        array has three dimensions and the unit axis is the zeroth one. This is intended solely to
        remove the dummy dimension that is in raw data from the summit.

        Setting `auto_squeeze = False` when initializing this object will never squeeze out any dimensions

        Returns
        -------
        data array
        """
        # This conditional is explicitly to catch summit data with a dummy first axis for WCS
        # purposes
        if self.auto_squeeze and len(self._hdu.data.shape) == 3 and self._hdu.data.shape[0] == 1:
            return np.squeeze(self._hdu.data)
        return self._hdu.data

    @data.setter
    def data(self, array: np.ndarray) -> None:
        # There is no shape magic stuff going on here right now because the tasks/services that care about
        # it will deal with it themselves (I think (tm)).
        self._hdu.data = array

    @property
    def header(self) -> fits.Header:
        return self._hdu.header

    @classmethod
    def from_header(
        cls, header: Union[fits.Header, dict], name: Optional[str] = None
    ) -> FitsAccessBase:
        """
        Convert a header to a CommonFitsData (or child) object

        Parameters
        ----------
        header
            A single `astropy.io.fits.header.Header` HDU object.
        name
            A unique name for the fits access instance
        """
        if isinstance(header, dict):
            header = fits.Header(header)
        hdu = fits.PrimaryHDU(header=header)
        return cls(hdu=hdu, name=name)



"""
Mixin for a WorkflowDataTaskBase subclass which implements input data set access functionality
"""
import json
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from dkist_processing_common.models.tags import Tag


@dataclass
class InputDatasetParameterValue:
    parameter_value_id: int
    parameter_value: Any = None
    parameter_value_start_date: Optional[datetime] = None


class InputDatasetMixin:
    """
    Mixin for WorkflowDataTaskBase that does x
    """

    @property
    def input_dataset_document(self):
        result = dict()
        paths: List[Path] = list(self.read(tags=[Tag.input_dataset()]))
        if not paths:
            return result
        if len(paths) > 1:
            raise ValueError("There are more than one input datasets to parse")
        p = paths[0]  # can loop in the future if multiple input datasets happen
        with p.open(mode="rb") as f:
            result = json.load(f)
        return result

    @property
    def input_dataset_frames(self) -> List[str]:
        return self.input_dataset_document.get("frames", list())

    @property
    def input_dataset_bucket(self) -> Union[str, None]:
        return self.input_dataset_document.get("bucket")

    @property
    def input_dataset_parameters(self) -> Dict[str, List[InputDatasetParameterValue]]:
        parameters = self.input_dataset_document.get("parameters", list())
        result = dict()
        for p in parameters:
            result.update(self._input_dataset_parse_parameter(p))
        return result

    def input_dataset_parameters_get(
        self,
        parameter_name: str,
        start_date: Optional[datetime] = None,
        default: Optional[Any] = None,
    ) -> Any:
        """ Get a single value from the input_dataset_parameters """
        start_date = start_date or datetime.utcnow()
        values = self.input_dataset_parameters.get(parameter_name, default)
        if values == default:
            return values
        sorted_values_from_before = sorted(
            [v for v in values if v.parameter_value_start_date <= start_date],
            key=lambda x: x.parameter_value_start_date,
        )
        try:
            result = sorted_values_from_before.pop().parameter_value
        except IndexError:
            raise ValueError(
                f"{parameter_name} has no values before {start_date.isoformat()} ({len(values)} values in total)"
            )
        return result

    def _input_dataset_parse_parameter(
        self, parameter: dict
    ) -> Dict[str, List[InputDatasetParameterValue]]:
        name: str = parameter["parameterName"]
        raw_values: List[dict] = parameter["parameterValues"]
        values = self._input_dataset_parse_parameter_values(raw_values=raw_values)
        return {name: values}

    def _input_dataset_parse_parameter_values(
        self, raw_values: List[Dict[str, Any]]
    ) -> List[InputDatasetParameterValue]:
        values = list()
        for v in raw_values:
            parsed_value = InputDatasetParameterValue(parameter_value_id=v["parameterValueId"])
            parsed_value.parameter_value = self._input_dataset_parse_parameter_value(
                raw_parameter_value=v["parameterValue"]
            )
            if d := v.get("parameterValueStartDate"):
                parsed_value.parameter_value_start_date = datetime.fromisoformat(d)
            values.append(parsed_value)
        return values

    @staticmethod
    def _input_dataset_parse_parameter_value(raw_parameter_value: str) -> Any:
        return json.loads(raw_parameter_value)



from pathlib import Path

import pytest

from dkist_processing_common.tasks.l1_output_data import L1OutputDataBase


class L1OutputDataBaseTask(L1OutputDataBase):
    def run(self) -> None:
        ...


@pytest.fixture
def l1_output_data_base_task(recipe_run_id):
    proposal_id = "test_proposal_id"
    with L1OutputDataBaseTask(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        task.constants["PROPOSAL_ID"] = proposal_id
        yield task, proposal_id
        task.constants.purge()


def test_format_object_key(l1_output_data_base_task):
    """
    :Given: a task based on L1OutputDataBase with a proposal ID in its constants mapping
    :When: formatting a path into an object key
    :Then: the proposal ID and filename are in the object key
    """
    task, proposal_id = l1_output_data_base_task
    filename = "test_filename.ext"
    path = Path(f"a/b/c/d/{filename}")
    assert proposal_id in task.format_object_key(path)
    assert filename in task.format_object_key(path)
    assert task.destination_bucket == "data"



"""
Extension of the GraphQL supporting retries for data processing use cases
"""
import logging

import requests
from gqlclient.base import GraphQLClientBase
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

logger = logging.getLogger(__name__)


class GraphQLClient(GraphQLClientBase):
    """
    Helper class for formatting and executing synchronous GraphQL queries and mutations
    """

    adapter = HTTPAdapter(
        max_retries=Retry(
            total=5,
            status_forcelist=[502, 503, 404],
            method_whitelist=["POST"],
        )
    )

    def execute_gql_call(self, query: dict, **kwargs) -> dict:
        """
        Executes a GraphQL query or mutation using requests.

        :param query: Dictionary formatted graphql query

        :param kwargs: Optional arguments that `requests` takes. e.g. headers

        :return: Dictionary containing the response from the GraphQL endpoint
        """
        logger.debug(f"Executing graphql call: host={self.gql_uri}")
        with requests.sessions.Session() as http:
            http.mount("http://", self.adapter)
            response = http.post(url=self.gql_uri, json=query, **kwargs)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            logger.error(
                f"Error executing graphql call: status_code={e.response.status_code}, detail={e.response.text}"
            )
            raise e
        return response.json()



"""
Scratch file system api
"""
from contextlib import contextmanager
from os import umask
from pathlib import Path
from shutil import rmtree
from typing import Generator
from typing import Iterable
from typing import Optional
from typing import Union

from dkist_processing_common._util.config import get_config
from dkist_processing_common._util.tags import TagDB


class WorkflowFileSystem:
    """
    Wrapper for interactions with the shared file system "scratch" supporting
    recipe run id based namespaces and tagged data
    """

    def __init__(
        self,
        recipe_run_id: Optional[int] = 0,
        task_name: str = "dev_task",
        scratch_base_path: Union[Path, str, None] = None,
    ):
        self.recipe_run_id = recipe_run_id
        self.task_name = task_name
        if not scratch_base_path:
            scratch_base_path = get_config("SCRATCH_BASE_PATH", "scratch/")
        self.scratch_base_path = scratch_base_path
        self.workflow_base_path = Path(self.scratch_base_path) / str(recipe_run_id)
        with self._mask():
            self.workflow_base_path.mkdir(parents=True, exist_ok=True)
        self._tag_db = TagDB(recipe_run_id=self.recipe_run_id, task_name=self.task_name)

    @staticmethod
    @contextmanager
    def _mask():
        """
        Set a permissive umask to allow other users (e.g. globus) to modify resources created
        by the scratch library
        """
        old_mask = umask(0)
        try:
            yield
        finally:
            umask(old_mask)

    def absolute_path(self, relative_path: Union[Path, str]) -> Path:
        """
        Convert a relative path to an absolute path with the base directories for the that workflow
        instance
        """
        relative_path = Path(relative_path)
        if relative_path.is_absolute():
            raise ValueError("Relative path must be relative")

        return self.workflow_base_path / relative_path

    def write(
        self,
        file_obj: bytes,
        relative_path: Union[Path, str],
        tags: Union[Iterable[str], None] = None,
    ) -> None:
        """
        Write a file object to the path specified and tagged with any tags listed in tags
        """
        path = self.absolute_path(relative_path)
        with self._mask():
            path.parent.mkdir(parents=True, exist_ok=True)
            with path.open(mode="wb") as f:
                f.write(file_obj)
        if tags:
            self.tag(path, tags)

    def delete(self, path: Union[Path, str]):
        path = Path(path)
        path.unlink(missing_ok=True)
        self._tag_db.clear_value(value=path)

    def tag(self, path: Union[Path, str], tags: Union[Iterable[str], str]) -> None:
        """
        Tag existing paths.  The path must be relative to the WorkflowFileSystem base path and
           must exist.
        """
        path = Path(path)
        if not (self.workflow_base_path in path.parents):
            raise ValueError(
                f"Cannot tag paths which are not children of the base path {self.workflow_base_path}"
            )
        if not path.exists():
            raise FileNotFoundError(f"Cannot tag paths which do not exist. {path=}")

        if isinstance(tags, str):
            self._tag_db.add(tags, str(path))
        else:
            for tag in tags:
                self._tag_db.add(tag, str(path))

    def tags(self, path: Union[Path, str]):
        value = str(path)
        return self._tag_db.tags_for_value(value=value)

    def find_any(self, tags: Iterable[str]) -> Generator[Path, None, None]:
        """
        Return a generator of Path objects that are tagged by the union of the input tags
        """
        for path in self._tag_db.any(tags):
            yield Path(path)

    def find_all(self, tags: Iterable[str]) -> Generator[Path, None, None]:
        """
        Return a generator of Path objects that are tagged by the intersection of the input tags
        """
        for path in self._tag_db.all(tags):
            yield Path(path)

    def close(self):
        """
        Close the db connection.  Call on __exit__ of a Task
        """
        self._tag_db.close()

    def purge(self, ignore_errors: bool = False):
        """
        Remove all data (tags, files, and folders) for the instance.
            Call when tearing down a workflow
        """
        rmtree(self.workflow_base_path, ignore_errors=ignore_errors)
        self._tag_db.purge()

    def __repr__(self):
        return f"WorkflowFileSystem(recipe_run_id={self.recipe_run_id}, task_name={self.task_name}, scratch_base_path={self.scratch_base_path})"

    def __str__(self):
        return f"{self!r} connected to {self._tag_db}"



"""
Common task to populate pipeline Constants and group files with tags by scanning headers
"""
import json
import logging
from typing import List
from typing import Tuple
from typing import TypeVar

from dkist_processing_common.models.constants import BudName
from dkist_processing_common.models.flower_pot import FlowerPot
from dkist_processing_common.models.flower_pot import Stem
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.parsers.cs_step import CSStepFlower
from dkist_processing_common.parsers.cs_step import NumCSStepBud
from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess
from dkist_processing_common.parsers.proposal_id_bud import ProposalIdBud
from dkist_processing_common.parsers.time import AverageCadenceBud
from dkist_processing_common.parsers.time import MaximumCadenceBud
from dkist_processing_common.parsers.time import MinimumCadenceBud
from dkist_processing_common.parsers.time import TimeOrderBud
from dkist_processing_common.parsers.time import VarianceCadenceBud
from dkist_processing_common.parsers.unique_bud import UniqueBud
from dkist_processing_common.tasks.base import ScienceTaskL0ToL1Base
from dkist_processing_common.tasks.mixin.input_dataset import InputDatasetMixin


__all__ = ["ParseL0InputData"]


logger = logging.getLogger(__name__)
S = TypeVar("S", bound=Stem)


class ParseL0InputData(ScienceTaskL0ToL1Base, InputDatasetMixin):
    @property
    def constant_flowers(self) -> List[S]:
        return [
            NumCSStepBud(self.input_dataset_parameters_get("MAX_CS_STEP_TIME_SEC")),
            UniqueBud(constant_name=BudName.instrument.value, metadata_key="instrument"),
            ProposalIdBud(),
            AverageCadenceBud(),
            MaximumCadenceBud(),
            MinimumCadenceBud(),
            VarianceCadenceBud(),
            TimeOrderBud(),
        ]

    @property
    def tag_flowers(self) -> List[S]:
        return [
            CSStepFlower(
                max_cs_step_time_sec=self.input_dataset_parameters_get("MAX_CS_STEP_TIME_SEC")
            ),
        ]

    @property
    def fits_parsing_class(self):
        return L0FitsAccess

    def run(self) -> None:

        with self.apm_step("Check that input frames exist"):
            self.check_input_frames()

        with self.apm_step("Ingest all input files"):
            tag_pot, constant_pot = self.make_flower_pots()

        with self.apm_step("Update constants"):
            self.update_constants(constant_pot)

        with self.apm_step("Tag files"):
            self.tag_petals(tag_pot)

    def make_flower_pots(self) -> Tuple[FlowerPot, FlowerPot]:
        """ Ingest all headers """
        tag_pot = FlowerPot()
        constant_pot = FlowerPot()
        tag_pot.flowers += self.tag_flowers
        constant_pot.flowers += self.constant_flowers

        for fits_obj in self.input_frames:
            filepath = fits_obj.name
            tag_pot.add_dirt(filepath, fits_obj)
            constant_pot.add_dirt(filepath, fits_obj)

        return tag_pot, constant_pot

    @property
    def input_frames(self):
        return list(
            self.fits_data_read_fits_access(
                tags=[Tag.input(), Tag.frame()], cls=self.fits_parsing_class
            )
        )

    def check_input_frames(self):
        """
        Helper function to make sure that at least one tagged frame exists before doing anything else
        """
        if len(self.input_frames) == 0:
            raise ValueError("No frames were tagged with INPUT and FRAME")

    def update_constants(self, constant_pot: FlowerPot):
        """ Update pipeline Constants """
        for flower in constant_pot:
            try:
                # If the value is a set, sort it before storing in redis
                if isinstance(flower.bud.value, set):
                    sorted_value = json.dumps(sorted(flower.bud.value))
                    self.constants.update({flower.stem_name: sorted_value})
                else:
                    self.constants.update({flower.stem_name: flower.bud.value})
            except StopIteration:
                # There are no petals
                pass

    def tag_petals(self, tag_pot: FlowerPot):
        """ Apply tags to file paths """
        for flower in tag_pot:
            for petal in flower.petals:
                tag = Tag.format_tag(flower.stem_name, petal.value)
                for path in petal.keys:
                    self.tag(path, tag)



"""
Tests for the workflow file system wrapper
"""
from pathlib import Path
from typing import Callable
from uuid import uuid4

import pytest

from dkist_processing_common._util.scratch import WorkflowFileSystem


@pytest.fixture()
def tagged_workflow_fs_data(workflow_file_system):
    wkflow_fs, _, _ = workflow_file_system
    file_obj = b"file contents"
    tags = ["A", "B", "C"]
    intersection = Path("Intersect/f.txt")
    union = [Path(f"Union{t}/f.txt") for t in tags]
    union.append(intersection)
    wkflow_fs.write(file_obj, intersection, tags=tags)
    for t, p in zip(tags, union):
        wkflow_fs.write(file_obj, p, tags=t)

    # prepend base path
    intersection = wkflow_fs.workflow_base_path / intersection
    union = [wkflow_fs.workflow_base_path / u for u in union]

    return tags, intersection, union, file_obj


def test_workflow_file_system(workflow_file_system):
    """
    Given: An instance of WorkflowFileSystem
    When: Accessing attributes
    Then: has attr workflow_base_path which is namespace(d) to recipe run id that exists
    """
    wkflow_fs, rrun_id, custom_base_path = workflow_file_system
    assert wkflow_fs.workflow_base_path.stem == str(rrun_id)
    assert wkflow_fs.workflow_base_path.exists()


@pytest.mark.parametrize(
    "folder_path",
    [
        pytest.param(Path("test_write_path"), id="Path"),
        pytest.param("test_write_str", id="str"),
    ],
)
@pytest.mark.parametrize(
    "tags",
    [
        pytest.param(None, id="0"),
        pytest.param("foo", id="1"),
        pytest.param(["foo", "baz"], id="2"),
    ],
)
def test_workflow_file_system_write(workflow_file_system, tags, folder_path):
    """
    Given: An instance of WorkflowFileSystem and a file to write
    When: writing to a relative path with/without tags
    Then: file is written to the path relative to the workflow fs configuration
      tags are added to the tag db if they exist
    """
    wkflow_fs, _, _ = workflow_file_system
    file_obj = uuid4().hex.encode("utf8")
    file_path = Path(f"{uuid4().hex[:6]}.bin")
    rel_path = folder_path / file_path
    # When
    wkflow_fs.write(file_obj, rel_path, tags=tags)
    # Then
    full_file_path = wkflow_fs.workflow_base_path / rel_path
    assert full_file_path.exists()
    with full_file_path.open(mode="rb") as f:
        assert file_obj == f.read()
    if tags:
        assert next(wkflow_fs.find_all(tags=tags)) == full_file_path


def test_workflow_file_system_write_invalid(workflow_file_system):
    """
    Given: An instance of WorkflowFileSystem
    When: writing to an absolute path
    Then: ValueError is raised
    """
    wkflow_fs, _, _ = workflow_file_system
    with pytest.raises(ValueError):
        wkflow_fs.write(b"1234", Path.cwd())


def test_workflow_file_system_tag(workflow_file_system):
    """
    Given: An instance of WorkflowFileSystem and a file already in a path
      relative to WorkflowFileSystem base path
    When: Tagging the path
    Then: Tag is added
    """
    wkflow_fs, _, _ = workflow_file_system
    path = wkflow_fs.workflow_base_path / Path("tag_test.txt")
    path.touch()
    # When
    wkflow_fs.tag(path, tags="tag_test")
    wkflow_fs.tag(str(path), tags="tag_test2")

    # Then
    assert next(wkflow_fs.find_any(tags="tag_test")) == path


def test_workflow_file_system_tag_invalid_base_path(workflow_file_system):
    """
    Given: An instance of WorkflowFileSystem
    When: tagging a path that isn't relative to the WorkflowFileSystem Base path
    Then: get a value error
    """
    wkflow_fs, _, _ = workflow_file_system
    bad_path = Path.cwd()
    with pytest.raises(ValueError):
        wkflow_fs.tag(bad_path, tags="bad_base")


def test_workflow_file_system_tag_invalid_path_not_exists(workflow_file_system):
    """
    Given: An instance of WorkflowFileSystem
    When: tagging a path that is relative to the WorkflowFileSystem base path
       doesn't exist
    Then: get a FileNotFoundError error
    """
    wkflow_fs, _, _ = workflow_file_system
    bad_path = wkflow_fs.workflow_base_path / Path("foo/bar.txt")
    assert not bad_path.exists()
    with pytest.raises(FileNotFoundError):
        wkflow_fs.tag(bad_path, tags="bad_base")


def test_workflow_file_system_find_any(workflow_file_system, tagged_workflow_fs_data):
    """
    Given: An instance of WorkflowFileSystem with tagged data
    When: Calling find any
    Then: Receive the union of the tagged data as Path objects to the data
    """
    tags, _, union, file_obj = tagged_workflow_fs_data
    wkflow_fs, _, _ = workflow_file_system

    # When
    result = list(wkflow_fs.find_any(tags))
    # Then
    assert len(result) == len(union)
    for path in result:
        assert path.exists()
        assert path in union
        with path.open(mode="rb") as f:
            assert f.read() == file_obj


def test_workflow_file_system_find_all(workflow_file_system, tagged_workflow_fs_data):
    """
    Given: An instance of WorkflowFileSystem with tagged data
    When: Calling find all
    Then: Receive the intersection of the tagged data as Path objects to the data
    """
    tags, intersection, _, file_obj = tagged_workflow_fs_data
    wkflow_fs, _, _ = workflow_file_system

    # When
    result = list(wkflow_fs.find_all(tags))
    # Then
    assert len(result) == 1
    path = result[0]
    assert path.exists()
    assert path == intersection
    with path.open(mode="rb") as f:
        assert f.read() == file_obj


def test_workflow_file_system_downstream_task(workflow_file_system, tagged_workflow_fs_data):
    """
    Given: An instance of WorkflowFileSystem with tagged data
    When: closing that instance and creating a new one
    Then: Receive the intersection of the tagged data as Path objects that were added
        to the original instance
    """
    wkflow_fs, rrun_id, custom_base_path = workflow_file_system
    tags, intersection, _, file_obj = tagged_workflow_fs_data
    # wkflow_fs.close()
    wkflow_fs2 = WorkflowFileSystem(
        recipe_run_id=rrun_id,
        task_name="wkflow_fs_test",
        scratch_base_path=custom_base_path,
    )
    # When
    try:
        result = list(wkflow_fs2.find_all(tags))
        # Then
        assert len(result) == 1
        path = result[0]
        assert path.exists()
        assert path == intersection
        with path.open(mode="rb") as f:
            assert f.read() == file_obj
    finally:
        # Teardown
        wkflow_fs2.purge()


def test_workflow_file_system_purge(workflow_file_system, tagged_workflow_fs_data):
    """
    Given: An instance of WorkflowFileSystem with tagged data
    When: calling purge
    Then: There is no more tagged data or paths in the base_path
    """
    tags, intersection, _, file_obj = tagged_workflow_fs_data
    wkflow_fs, _, _ = workflow_file_system

    # When
    wkflow_fs.purge()
    # Then
    result = list(wkflow_fs.find_any(tags))
    assert not result
    assert not wkflow_fs.workflow_base_path.exists()


@pytest.mark.parametrize(
    "func, attr",
    [
        pytest.param(repr, "__repr__", id="repr"),
        pytest.param(str, "__str__", id="str"),
    ],
)
def test_workflow_file_system_dunder(workflow_file_system, func: Callable, attr):
    """
    Given: Connection to a tag database
    When: retrieving dunder method that should be implemented
    Then: It is implemented
    """
    wkflow_fs, _, _ = workflow_file_system

    assert getattr(wkflow_fs, attr, None)
    assert func(wkflow_fs)


def test_workflow_file_system_delete(workflow_file_system, tagged_workflow_fs_data):
    """
    Given: An instance of WorkflowFileSystem with tagged data
    When: Deleting a specific file
    Then: The file is removed from disk and all references to it are gone from tags
    """
    tags, _, union, file_obj = tagged_workflow_fs_data
    wkflow_fs, _, _ = workflow_file_system
    assert union[0].exists()
    assert str(union[0]) in wkflow_fs._tag_db.any(tags=tags)
    wkflow_fs.delete(path=union[0])
    assert not union[0].exists()
    assert not str(union[0]) in wkflow_fs._tag_db.any(tags=tags)



import logging
import random

import numpy as np
import pytest

from dkist_processing_common.parsers.cs_step import CSStep


def test_equal_correct(grouped_cal_sequence_headers):
    """
    Given: A set of PolCal headers
    When: Converting them to CSStep objects and comparing them
    Then: All headers belonging to the same step produce CSStep objects that are equal
    """
    for cs_header_list in grouped_cal_sequence_headers.values():
        for i in range(len(cs_header_list)):
            assert CSStep(cs_header_list[0]) == CSStep(cs_header_list[i])


@pytest.mark.parametrize("stepnum", [pytest.param(i, id=f"step {i}") for i in range(7)])
def test_not_equal_correct(grouped_cal_sequence_headers, stepnum):
    """
    Given: A set of PolCal headers
    When: Converting them to CSStep objects and comparing them
    Then: All objects from a single step are not equal to all objects from other steps
    """
    for i in range(7):
        if i != stepnum:
            assert CSStep(grouped_cal_sequence_headers[stepnum][0]) != CSStep(
                grouped_cal_sequence_headers[i][0]
            )


def test_not_equal_non_CS_Step_type(grouped_cal_sequence_headers):
    """
    Given: A PolCal header and resulting CSStep object
    When: Testing equality with a non CSStep object
    Then: An error is raised
    """
    cs_step = CSStep(grouped_cal_sequence_headers[0][0])
    with pytest.raises(TypeError):
        _ = cs_step == 1


def test_order_correct(grouped_cal_sequence_headers):
    """
    Given: A set of PolCal headers
    When: Converting them to CSStep objects and ordering them
    Then: The step objects are correctly ordered by observe time
    """
    logging.warning(f"{grouped_cal_sequence_headers}")
    cs_step_list = [CSStep(header) for header in sum(grouped_cal_sequence_headers.values(), [])]
    random.shuffle(cs_step_list)  # Just to mix it up a bit
    time_list = [c.obs_time for c in cs_step_list]
    cs_sort_idx = np.argsort(cs_step_list)
    time_sort_idx = np.argsort(time_list)
    assert np.array_equal(cs_sort_idx, time_sort_idx)



import json
from pathlib import Path
from uuid import uuid4

import pytest

from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.tasks.transfer_input_data import TransferL0Data


@pytest.fixture
def transfer_l0_data(recipe_run_id, tmp_path):
    bucket_name = "transfer_l0_bucket"
    task = TransferL0Data(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    )
    task.scratch = WorkflowFileSystem(
        recipe_run_id=recipe_run_id,
        scratch_base_path=tmp_path,
    )
    task.scratch.scratch_base_path = tmp_path
    file_path = task.scratch.workflow_base_path / Path(f"{uuid4().hex[:6]}.ext")
    input_dataset = {
        "bucket": bucket_name,
        "parameters": [
            {
                "parameterName": "param_name",
                "parameterValues": [
                    {
                        "parameterValueId": 1,
                        "parameterValue": [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
                        "parameterValueStartDate": "2000-01-01",
                    }
                ],
            }
        ],
        "frames": [
            Path("objectkey1").as_posix(),
            Path("objectkey2").as_posix(),
            Path("objectkey3").as_posix(),
        ],
    }
    file_path.write_text(data=json.dumps(input_dataset))
    task.tag(path=file_path, tags=Tag.input_dataset())
    yield task, input_dataset
    task.scratch.purge()


def test_format_transfer_items(transfer_l0_data):
    """
    :Given: a TransferL0Data task with a valid input dataset
    :When: formatting items in the input dataset for transfer
    :Then: the items are correctly loaded into GlobusTransferItem objects
    """
    task, input_dataset = transfer_l0_data
    filenames = [Path(frame).name for frame in input_dataset["frames"]]
    assert len(task.format_transfer_items()) == len(input_dataset["frames"])
    for item in task.format_transfer_items():
        assert item.source_path.as_posix() in [
            "/" + input_dataset["bucket"] + "/" + frame for frame in input_dataset["frames"]
        ]
        assert item.destination_path.name in filenames
        assert not item.recursive



"""
Data structures for messages placed on the interservice bus
"""
from talus.message import message_class


@message_class(routing_key="catalog.frame.m", queues=["catalog.frame.q"])
class CatalogFrameMessage:
    """
    Class to hold the catalog_frame_message to be sent to RabbitMQ
    """

    objectName: str
    conversationId: str
    bucket: str
    incrementDatasetCatalogReceiptCount: bool = True


@message_class(routing_key="catalog.object.m", queues=["catalog.object.q"])
class CatalogObjectMessage:
    """
    Class to hold the catalog_object_message to be sent to RabbitMQ
    """

    objectType: str
    objectName: str
    conversationId: str
    bucket: str
    groupId: str = "default_group_id"
    groupName: str = "DATASET"
    incrementDatasetCatalogReceiptCount: bool = True



"""
Mixin for a WorkflowDataTaskBase subclass which implements Object Store data access functionality
"""
import json
from os import environ
from pathlib import Path
from typing import Union

from object_clerk import ObjectClerk

from dkist_processing_common._util.config import get_mesh_config


class ObjectStoreMixin:
    @property
    def _object_store_retry_configuration(self) -> dict:
        retry_configuration = environ.get(
            "RETRY_CONFIG",
            '{"retry_delay":1,"retry_backoff":2,"retry_jitter":[1,10],"retry_max_delay":300}',
        )
        retry_configuration = json.loads(retry_configuration)
        retry_configuration["retry_jitter"] = tuple(retry_configuration["retry_jitter"])
        return retry_configuration

    @property
    def _object_store_connection_configuration(self) -> dict:
        mesh_config = get_mesh_config()
        conn_config = {
            "host": mesh_config["object-store-api"]["mesh_address"],
            "port": mesh_config["object-store-api"]["mesh_port"],
            "access_key": environ.get("OBJECT_STORE_ACCESS_KEY", None),
            "secret_key": environ.get("OBJECT_STORE_SECRET_KEY", None),
            "use_ssl": False,
        }
        return conn_config

    @property
    def object_store_client(self) -> ObjectClerk:
        """
        The object store client is additionally configured by the following environment variables:
          MULTIPART_THRESHOLD - Threshold in bytes at which uploads are broken into multiple parts
            for upload. Impacts the checksum stored in the eTag
          S3_CLIENT_CONFIG - https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
          S3_UPLOAD_CONFIG - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/customizations/s3.html#boto3.s3.transfer.TransferConfig
        """
        retry_config = self._object_store_retry_configuration
        conn_config = self._object_store_connection_configuration
        return ObjectClerk(**conn_config, **retry_config)

    def object_store_upload_movie(
        self,
        movie: Union[Path, bytes],
        bucket: str,
        object_key: str,
        content_type: str = "video/mp4",
    ):
        self.object_store_client.upload_object(
            object_data=movie,
            bucket=bucket,
            object_key=object_key,
            verify_checksum=True,
            content_type=content_type,
            group_name="DATASET",
            group_id=self.dataset_id,
            object_type="MOVIE",
        )



"""
Task(s) for the transfer out of data from a processing pipeline
"""
import logging
from abc import ABC
from pathlib import Path
from typing import Iterable
from typing import List

from dkist_processing_common.models.message import CatalogFrameMessage
from dkist_processing_common.models.message import CatalogObjectMessage
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.tasks.base import ParsedL0InputTaskBase
from dkist_processing_common.tasks.mixin.globus import GlobusMixin
from dkist_processing_common.tasks.mixin.globus import GlobusTransferItem
from dkist_processing_common.tasks.mixin.interservice_bus import InterserviceBusMixin
from dkist_processing_common.tasks.mixin.metadata_store import MetadataStoreMixin
from dkist_processing_common.tasks.mixin.object_store import ObjectStoreMixin


__all__ = ["AddDatasetReceiptAccount", "PublishCatalogMessages", "TransferL1Data"]


logger = logging.getLogger(__name__)


class L1OutputDataBase(ParsedL0InputTaskBase, ABC):
    destination_bucket: str = "data"

    def format_object_key(self, path: Path) -> str:
        """
        Convert output paths into object store keys
        Parameters
        ----------
        path: the Path to convert

        Returns
        -------
        formatted path in the object store
        """
        return str(Path(self.proposal_id, self.dataset_id, path.name))


class TransferL1Data(L1OutputDataBase, GlobusMixin, ObjectStoreMixin):
    """
    Transfers Level 1 processed data to the object store
    """

    def transfer_science_frames(self):
        transfer_items = []
        for file_path in self.read(tags=[Tag.output(), Tag.frame()]):
            object_key = self.format_object_key(file_path)
            destination_path = Path(self.destination_bucket, object_key)
            item = GlobusTransferItem(
                source_path=file_path,
                destination_path=destination_path,
            )
            transfer_items.append(item)
        logger.info(
            f"Preparing globus transfer {len(transfer_items)} items: recipe_run_id={self.recipe_run_id}. transfer_items={transfer_items[:3]}..."
        )
        self.globus_transfer_scratch_to_object_store(
            transfer_items=transfer_items,
            label=f"Transfer Output Data for recipe_run_id {self.recipe_run_id}",
        )

    def transfer_movie(self):
        paths = list(self.read(tags=[Tag.output(), Tag.movie()]))
        if len(paths) == 0:
            logger.warning(
                f"No movies found to upload for dataset. recipe_run_id={self.recipe_run_id}"
            )
            return
        movie = paths[0]
        if count := len(paths) > 1:
            # note: this needs to be an error or the dataset receipt accounting will have an
            # expected count > the eventual actual
            raise RuntimeError(
                f"Multiple movies found to upload.  Uploading the first one. "
                f"{count=}, {movie=}, recipe_run_id={self.recipe_run_id}"
            )
        logger.info(f"Uploading Movie: recipe_run_id={self.recipe_run_id}, {movie=}")
        movie_object_key = self.format_object_key(movie)
        self.object_store_upload_movie(
            movie=movie,
            bucket=self.destination_bucket,
            object_key=movie_object_key,
            content_type="video/mp4",
        )

    def run(self) -> None:
        with self.apm_step("Upload Science Frames"):
            self.transfer_science_frames()
        with self.apm_step("Upload Movie"):
            self.transfer_movie()


class PublishCatalogMessages(L1OutputDataBase, InterserviceBusMixin):
    def frame_messages(self, paths: Iterable[Path]) -> List[CatalogFrameMessage]:
        messages = [
            CatalogFrameMessage(
                objectName=self.format_object_key(path=p),
                conversationId=str(self.recipe_run_id),
                bucket=self.destination_bucket,
            )
            for p in paths
        ]
        return messages

    def object_messages(self, paths: Iterable[Path], object_type: str):
        messages = [
            CatalogObjectMessage(
                objectType=object_type,
                objectName=self.format_object_key(p),
                bucket=self.destination_bucket,
                conversationId=str(self.recipe_run_id),
                groupId=self.dataset_id,
            )
            for p in paths
        ]
        return messages

    def run(self) -> None:
        with self.apm_step("Gather output data"):
            frames = self.read(tags=[Tag.output(), Tag.frame()])
            movies = self.read(tags=[Tag.output(), Tag.movie()])
        with self.apm_step("Create message objects"):
            messages = []
            messages += self.frame_messages(paths=frames)
            messages += self.object_messages(paths=movies, object_type="MOVIE")
        with self.apm_step("Publish messages"):
            self.interservice_bus_publish(messages=messages)


class AddDatasetReceiptAccount(ParsedL0InputTaskBase, MetadataStoreMixin):
    def run(self) -> None:
        with self.apm_step("Count Expected Outputs"):
            dataset_id = self.dataset_id
            expected_object_count = len(list(self.read(tags=Tag.output())))
        logger.info(
            f"Adding Dataset Receipt Account: "
            f"{dataset_id=}, {expected_object_count=}, recipe_run_id={self.recipe_run_id}"
        )
        with self.apm_step("Add Dataset Receipt Account"):
            self.metadata_store_add_dataset_receipt_account(
                dataset_id=dataset_id, expected_object_count=expected_object_count
            )



"""
Mixin for a WorkflowDataTaskBase subclass which implements Metadata Store data access functionality
"""
from typing import Union

from dkist_processing_common._util.config import get_mesh_config
from dkist_processing_common._util.graphql import GraphQLClient
from dkist_processing_common.models.graphql import CreateRecipeRunStatusResponse
from dkist_processing_common.models.graphql import DatasetCatalogReceiptAccountMutation
from dkist_processing_common.models.graphql import RecipeRunMutation
from dkist_processing_common.models.graphql import RecipeRunProvenanceMutation
from dkist_processing_common.models.graphql import RecipeRunQuery
from dkist_processing_common.models.graphql import RecipeRunResponse
from dkist_processing_common.models.graphql import RecipeRunStatusMutation
from dkist_processing_common.models.graphql import RecipeRunStatusQuery
from dkist_processing_common.models.graphql import RecipeRunStatusResponse


class MetadataStoreMixin:
    @property
    def metadata_store_client(self) -> GraphQLClient:
        mesh_config = get_mesh_config()
        return GraphQLClient(
            f'http://{mesh_config["internal-api-gateway"]["mesh_address"]}:{mesh_config["internal-api-gateway"]["mesh_port"]}/graphql'
        )

    def metadata_store_change_recipe_run_to_inprogress(self):
        """
        Set the recipe run status to "INPROGRESS"
        """
        self._metadata_store_change_status(status="INPROGRESS", is_complete=False)

    def metadata_store_change_recipe_run_to_completed_successfully(self):
        """
        Set the recipe run status to "COMPLETEDSUCCESSFULLY"
        """
        self._metadata_store_change_status(status="COMPLETEDSUCCESSFULLY", is_complete=True)

    def metadata_store_add_dataset_receipt_account(
        self, dataset_id: str, expected_object_count: int
    ):
        self.metadata_store_client.execute_gql_mutation(
            mutation_base="createDatasetCatalogReceiptAccount",
            mutation_parameters=DatasetCatalogReceiptAccountMutation(
                datasetId=dataset_id, expectedObjectCount=expected_object_count
            ),
        )

    def metadata_store_record_provenance(self, is_task_manual: bool, library_versions: str):
        params = RecipeRunProvenanceMutation(
            inputDatasetId=self.metadata_store_input_dataset_id,
            isTaskManual=is_task_manual,
            recipeRunId=self.recipe_run_id,
            taskName=self.task_name,
            libraryVersions=library_versions,
            workflowVersion=self.workflow_version,
        )
        self.metadata_store_client.execute_gql_mutation(
            mutation_base="createRecipeRunProvenance", mutation_parameters=params
        )

    @property
    def metadata_store_input_dataset_document(self) -> str:
        return self._metadata_store_recipe_run().recipeInstance.inputDataset.inputDatasetDocument

    @property
    def metadata_store_input_dataset_id(self) -> int:
        return self._metadata_store_recipe_run().recipeInstance.inputDataset.inputDatasetId

    @property
    def metadata_store_recipe_instance_id(self) -> int:
        return self._metadata_store_recipe_run().recipeInstanceId

    @property
    def metadata_store_recipe_id(self) -> int:
        return self._metadata_store_recipe_run().recipeInstance.recipeId

    def _metadata_store_recipe_run(self, allow_cache: bool = True) -> RecipeRunResponse:
        is_cached = bool(getattr(self, "_recipe_run_cache", False))
        if is_cached and allow_cache:
            return self._recipe_run_cache
        response = self.metadata_store_client.execute_gql_query(
            query_base="recipeRuns",
            query_response_cls=RecipeRunResponse,
            query_parameters=RecipeRunQuery(recipeRunId=self.recipe_run_id),
        )
        self._recipe_run_cache = response[0]
        return self._recipe_run_cache

    def _metadata_store_change_status(self, status: str, is_complete: bool):
        """
        Change the recipe run status of a recipe run to the given status
        """
        recipe_run_status_id = self._metadata_store_recipe_run_status_id(status=status)
        if not recipe_run_status_id:
            recipe_run_status_id = self._metadata_store_create_recipe_run_status(
                status=status, is_complete=is_complete
            )
        self._metadata_store_update_status(recipe_run_status_id=recipe_run_status_id)

    def _metadata_store_recipe_run_status_id(self, status: str) -> Union[None, int]:
        """
        Find the id of a recipe run status
        """
        response = self.metadata_store_client.execute_gql_query(
            query_base="recipeRunStatuses",
            query_response_cls=RecipeRunStatusResponse,
            query_parameters=RecipeRunStatusQuery(recipeRunStatusName=status),
        )
        if len(response) > 0:
            return response[0].recipeRunStatusId

    def _metadata_store_create_recipe_run_status(self, status: str, is_complete: bool) -> int:
        """
        Add a new recipe run status to the db
        :param status: name of the status to add
        :param is_complete: does the new status correspond to an accepted completion state
        """
        recipe_run_statuses = {
            "INPROGRESS": "Recipe run is currently undergoing processing",
            "COMPLETEDSUCCESSFULLY": "Recipe run processing completed with no errors",
        }

        if not isinstance(status, str):
            raise TypeError(f"status must be of type str: {status}")
        if not isinstance(is_complete, bool):
            raise TypeError(f"is_complete must be of type bool: {is_complete}")
        recipe_run_status_response = self.metadata_store_client.execute_gql_mutation(
            mutation_base="createRecipeRunStatus",
            mutation_response_cls=CreateRecipeRunStatusResponse,
            mutation_parameters=RecipeRunStatusMutation(
                recipeRunStatusName=status,
                isComplete=is_complete,
                recipeRunStatusDescription=recipe_run_statuses[status],
            ),
        )
        return recipe_run_status_response.recipeRunStatus.recipeRunStatusId

    def _metadata_store_update_status(
        self,
        recipe_run_status_id: int,
    ):
        """
        Change the status of a given recipe run id
        :param recipe_run_status_id: the new status to use
        :param recipe_run_id: id of the recipe run to have the status changed
        """
        self.metadata_store_client.execute_gql_mutation(
            mutation_base="updateRecipeRun",
            mutation_parameters=RecipeRunMutation(
                recipeRunId=self.recipe_run_id, recipeRunStatusId=recipe_run_status_id
            ),
        )



import logging
from pathlib import Path
from typing import Callable
from typing import Optional

from dkist_processing_core.task import TaskBase

from dkist_processing_common.tasks.base import WorkflowDataTaskBase


class ManualProcessing:
    def __init__(
        self, workflow_path: Path, recipe_run_id: Optional[int] = 1, testing: bool = False
    ):
        self.workflow_path = workflow_path
        self.recipe_run_id = recipe_run_id
        self.testing = testing

    def run_task(self, task: Callable) -> None:
        """
        Wrapper function for calling the .run() method on a DKIST processing pipeline task
        Parameters
        ----------
        task: Callable
            task object that subclasses TaskBase
        Returns
        -------
        None
        """
        if not issubclass(task, TaskBase):
            raise RuntimeError(
                "Task is not a valid DKIST processing task. "
                "Must be a subclass of dkist_processing_core.task.TaskBase"
            )
        t = task(
            recipe_run_id=self.recipe_run_id, workflow_name="manual", workflow_version="manual"
        )
        t.scratch.scratch_base_path = self.workflow_path
        t.scratch.workflow_base_path = Path(t.scratch.scratch_base_path) / str(self.recipe_run_id)
        if self.testing:
            task.metadata_store_recipe_id = self.recipe_run_id + 1
            task.metadata_store_recipe_instance_id = self.recipe_run_id + 2
        t.run()
        logging.info(f"Task {task.__name__} completed")

    def purge_tags_and_constants(self) -> None:
        """
        Remove all filepath tags and constants from the associated objects.
        Run at the end of a manual processing run.
        Returns
        -------
        None
        """

        class PurgeTagsAndConstants(WorkflowDataTaskBase):
            def run(self):
                pass

        t = PurgeTagsAndConstants(
            recipe_run_id=self.recipe_run_id, workflow_name="manual", workflow_version="manual"
        )
        t.scratch._tag_db.purge()
        t.constants.purge()
        logging.info(f"Constants and filepath tags purged for recipe run id {self.recipe_run_id}")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.purge_tags_and_constants()



import pytest
from astropy.io import fits

from dkist_processing_common.models.fits_access import FitsAccessBase
from dkist_processing_common.parsers.cs_step import CSStepFlower
from dkist_processing_common.parsers.cs_step import NumCSStepBud
from dkist_processing_common.parsers.proposal_id_bud import ProposalIdBud
from dkist_processing_common.parsers.single_value_single_key_flower import (
    SingleValueSingleKeyFlower,
)
from dkist_processing_common.parsers.unique_bud import UniqueBud


class FitsReader(FitsAccessBase):
    def __init__(self, hdu, name):
        super().__init__(hdu, name)
        self.thing_id: int = self.header["id_key"]
        self.constant_thing: float = self.header["constant"]
        self.name = name
        self.proposal_id: str = self.header.get("ID___013")
        self.ip_task_type: str = self.header.get("DKIST004")


@pytest.fixture()
def basic_header_objs():
    header_dict = {
        "thing0": fits.header.Header(
            {"id_key": 0, "constant": 6.28, "DKIST004": "observe", "ID___013": "proposal_id_1"}
        ),
        "thing1": fits.header.Header(
            {"id_key": 1, "constant": 6.28, "DKIST004": "observe", "ID___013": "proposal_id_1"}
        ),
        "thing2": fits.header.Header(
            {"id_key": 2, "constant": 6.28, "DKIST004": "dark", "ID___013": "proposal_id_2"}
        ),
        "thing3": fits.header.Header(
            {"id_key": 0, "constant": 6.28, "DKIST004": "observe", "ID___013": "proposal_id_1"}
        ),
    }
    return (FitsReader.from_header(header, name=path) for path, header in header_dict.items())


@pytest.fixture()
def bad_header_objs():
    bad_headers = {
        "thing0": fits.header.Header({"id_key": 0, "constant": 6.28}),
        "thing1": fits.header.Header({"id_key": 1, "constant": 3.14}),
    }
    return (FitsReader.from_header(header, name=path) for path, header in bad_headers.items())


def test_unique_bud(basic_header_objs):
    """
    Given: A set of headers with a constant value header key
    When: Ingesting headers with a UniqueRock and asking for the value
    Then: The Rock's value is the header constant value
    """
    bud = UniqueBud(
        constant_name="constant",
        metadata_key="constant_thing",
    )
    assert bud.stem_name == "constant"
    for fo in basic_header_objs:
        key = fo.name
        bud.update(key, fo)

    petal = list(bud.petals)
    assert len(petal) == 1
    assert petal[0].value == 6.28


def test_unique_bud_non_unique_inputs(bad_header_objs):
    """
    Given: A set of headers with a non-constant header key that is expected to be constant
    When: Ingesting headers with a UniqueRock and asking for the value
    Then: An error is raised
    """
    rock = UniqueBud(
        constant_name="constant",
        metadata_key="constant_thing",
    )
    assert rock.stem_name == "constant"
    for fo in bad_header_objs:
        key = fo.name
        rock.update(key, fo)

    with pytest.raises(ValueError):
        assert next(rock.petals)


def test_single_value_single_key_flower(basic_header_objs):
    """
    Given: A set of filepaths and associated headers with a single key that has a limited set of values
    When: Ingesting with a SingleValueSingleKeyFlower and asking for the grouping
    Then: The filepaths are grouped correctly based on the header key value
    """
    flower = SingleValueSingleKeyFlower(tag_stem_name="id", metadata_key="thing_id")
    assert flower.stem_name == "id"
    for fo in basic_header_objs:
        key = fo.name
        flower.update(key, fo)

    petals = sorted(list(flower.petals), key=lambda x: x.value)
    assert len(petals) == 3
    assert petals[0].value == 0
    assert petals[0].keys == ["thing0", "thing3"]
    assert petals[1].value == 1
    assert petals[1].keys == ["thing1"]
    assert petals[2].value == 2
    assert petals[2].keys == ["thing2"]


def test_cs_step_flower(grouped_cal_sequence_headers, non_polcal_headers):
    """
    Given: A set of PolCal headers, non-PolCal headers, and the CSStepFlower
    When: Updating the CSStepFlower with all headers
    Then: The flower correctly organizes the PolCal frames and ignores the non-PolCal frames
    """
    cs_step_flower = CSStepFlower()
    for step, headers in grouped_cal_sequence_headers.items():
        for i, h in enumerate(headers):
            key = f"step_{step}_file_{i}"
            cs_step_flower.update(key, h)

    for h in non_polcal_headers:
        cs_step_flower.update("non_polcal", h)

    assert len(list(cs_step_flower.petals)) == len(list(grouped_cal_sequence_headers.keys()))
    for step_petal in cs_step_flower.petals:
        assert sorted(step_petal.keys) == [
            f"step_{step_petal.value}_file_{i}" for i in range(len(step_petal.keys))
        ]


def test_num_cs_step_bud(grouped_cal_sequence_headers, non_polcal_headers):
    """
    Given: A set of PolCal headers, non-PolCal headers, and the NumCSStepRock
    When: Updating the NumCSStepRock with all headers
    Then: The rock reports the correct number of CS Steps (thus ignoring the non-PolCal frames)
    """
    num_cs_bud = NumCSStepBud()
    for step, headers in grouped_cal_sequence_headers.items():
        for h in headers:
            num_cs_bud.update(step, h)

    for h in non_polcal_headers:
        num_cs_bud.update("foo", h)

    bud = list(num_cs_bud.petals)
    assert len(bud) == 1
    assert bud[0].value == len(grouped_cal_sequence_headers.keys())


def test_proposal_id_bud(basic_header_objs):
    bud = ProposalIdBud()
    assert bud.stem_name == "PROPOSAL_ID"
    for fo in basic_header_objs:
        key = fo.name
        bud.update(key, fo)

    petal = list(bud.petals)
    assert len(petal) == 1
    assert petal[0].value == "proposal_id_1"



import random
import string

import pytest


@pytest.fixture(scope="function")
def test_dict() -> dict:
    data = {
        "KEY 0": random.randint(-(2 ** 51), 2 ** 51),
        "KEY 1": random.random(),
        "KEY 2": "".join(
            [random.choice(string.ascii_letters) for _ in range(random.randint(0, 512))]
        ),
    }
    return data


def test_constants_as_dict(constants, test_dict):
    """
    Given: a Constants object and a python dictionary
    When: treating the Constants object as a dict for getting and setting
    Then: the Constants object behaves like a dictionary
    """
    for key, value in test_dict.items():
        constants[key] = value

    assert len(constants) == len(test_dict)
    assert sorted(list(constants)) == sorted(list(test_dict.keys()))
    for key, value in test_dict.items():
        assert constants[key] == value


def test_key_exists(constants):
    """
    Given: a populated Constants object
    When: trying to set a key that already exists
    Then: an error is raised
    """
    constants["foo"] = "baz"
    with pytest.raises(ValueError):
        constants["foo"] = "baz2: electric bogaloo"


def test_replace_key(constants):
    """
    Given: a populated Constants object
    When: a constant key is deleted
    Then: the key is removed and no longer exists in the Constants object
    """
    constants["foo"] = "baz"
    del constants["foo"]
    assert "foo" not in constants
    constants["foo"] = "baz3: tokyo drift"
    assert constants["foo"] == "baz3: tokyo drift"


def test_key_does_not_exist(constants):
    """
    Given: a Constants object
    When: trying to get a constant value that doesn't exist
    Then: an error is raised
    """
    with pytest.raises(KeyError):
        _ = constants["foo"]



"""
Mixin for a WorkflowDataTaskBase subclass which implements fits data retrieval functionality
"""
from io import BytesIO
from pathlib import Path
from typing import Generator
from typing import Iterable
from typing import Tuple
from typing import Type
from typing import Union

from astropy.io import fits

from dkist_processing_common.models.fits_access import FitsAccessBase

tag_type_hint = Union[Iterable[str], str]


class FitsDataMixin:
    """
    Mixin for the WorkflowDataTaskBase to support fits r/w operations
    """

    def fits_data_read(
        self, tags: tag_type_hint
    ) -> Generator[Tuple[Path, fits.HDUList], None, None]:
        for path in self.read(tags=tags):
            yield path, self.fits_data_open(path)

    def fits_data_read_hdu(
        self, tags: tag_type_hint
    ) -> Generator[Tuple[Path, Union[fits.PrimaryHDU, fits.CompImageHDU]], None, None]:
        for path, hdul in self.fits_data_read(tags=tags):
            yield path, self.fits_data_extract_hdu(hdul=hdul)

    def fits_data_read_fits_access(
        self, tags: tag_type_hint, cls: Type[FitsAccessBase]
    ) -> Generator[FitsAccessBase, None, None]:
        for path, hdu in self.fits_data_read_hdu(tags=tags):
            yield cls(hdu=hdu, name=str(path))

    @staticmethod
    def fits_data_extract_hdu(hdul: fits.HDUList) -> Union[fits.PrimaryHDU, fits.CompImageHDU]:
        if hdul[0].data is not None:
            return hdul[0]
        return hdul[1]

    @staticmethod
    def fits_data_open(path: Union[str, Path]) -> fits.HDUList:
        return fits.open(path)

    def fits_data_write(
        self,
        hdu_list: fits.HDUList,
        tags: tag_type_hint,
        relative_path: Union[Path, str, None] = None,
    ) -> Path:
        file_obj = BytesIO()
        hdu_list.writeto(file_obj, checksum=True)
        file_obj.seek(0)
        return self.write(file_obj=file_obj, tags=tags, relative_path=relative_path)



import numpy as np
import pytest
from astropy.io import fits

from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess


@pytest.fixture()
def hdu_with_complete_common_header(complete_common_header):
    """
    An HDU with data and a header with some common by-frame keywords and a single instrument
    specific one.
    """
    data = np.arange(9).reshape(3, 3)
    hdu = fits.PrimaryHDU(data, header=complete_common_header)
    return hdu


@pytest.fixture()
def extra_axis_hdu_with_complete_common_header(complete_common_header):
    """
    An HDU with data and a header with some common by-frame keywords and a single instrument
    specific one. Also contains an axis of length 1 in the data array
    """
    data = np.ones(shape=(1, 3, 3))
    hdu = fits.PrimaryHDU(data, header=complete_common_header)
    return hdu


@pytest.fixture()
def singleton_axis_hdu_with_complete_common_header(complete_common_header):
    """
    An HDU with data and a header with some common by-frame keywords and a single instrument
    specific one. Also contains an axis of length 1 in the data array
    """
    data = np.ones(shape=(1, 3))
    hdu = fits.PrimaryHDU(data, header=complete_common_header)
    return hdu


@pytest.fixture()
def three_singleton_axes_hdu(complete_common_header):
    """
    An HDU with data and a full header. The data has three dimensions but only a single value
    """
    data = np.ones(shape=(1, 1, 1))
    hdu = fits.PrimaryHDU(data, header=complete_common_header)
    return hdu


@pytest.fixture()
def hdu_with_no_data(complete_common_header):
    """
    An HDU with data and a header with some common by-frame keywords and a single instrument
    specific one.  No data is included in any HDUs.
    """
    hdu = fits.PrimaryHDU(header=complete_common_header)
    return hdu


@pytest.fixture()
def hdu_with_incomplete_common_header(tmp_path):
    """
    An HDU with data and a header missing one of the expected common by-frame keywords
    """
    data = np.arange(9).reshape(3, 3)
    hdu = fits.PrimaryHDU(data)
    hdu.header["TELEVATN"] = 6.28
    hdu.header["TAZIMUTH"] = 3.14
    return hdu


def test_from_single_hdu(hdu_with_complete_common_header):
    """
    Given: an HDU with expected, common by-frame keywords
    When: loading the HDU with the CommonFitsData class
    Then: all values for common keywords are exposed as properties on the fits_obj class
    """
    fits_obj = L0FitsAccess(hdu_with_complete_common_header)
    assert fits_obj.elevation == 6.28
    assert fits_obj.azimuth == 3.14
    assert fits_obj.table_angle == 1.23
    assert fits_obj.time_obs == "1988-05-25T01:23:45.678"
    assert fits_obj.name is None
    np.testing.assert_equal(fits_obj.data, np.arange(9).reshape(3, 3))


def test_from_header(hdu_with_complete_common_header):
    """
    Given: an HDU with expected, common by-frame keywords
    When: constructing a L0FitsAccess object via the .from_header method
    Then: all values for common keywords are exposed as properties on the fits_obj class
    """
    fits_obj = L0FitsAccess.from_header(hdu_with_complete_common_header.header)
    assert fits_obj.elevation == 6.28
    assert fits_obj.azimuth == 3.14
    assert fits_obj.table_angle == 1.23
    assert fits_obj.time_obs == "1988-05-25T01:23:45.678"
    assert fits_obj.name is None


def test_no_header_value(hdu_with_incomplete_common_header):
    """
    Given: an HDU with a header with missing common by-frame keywords
    When: processing the HDU with the CommonFitsData class
    Then: a KeyError is raised
    """
    with pytest.raises(KeyError):
        _ = L0FitsAccess(hdu_with_incomplete_common_header)


def test_as_subclass(hdu_with_complete_common_header):
    """
    Given: an instrument-specific fits_obj class that subclasses CommonFrameMetadata
    When: processing a HDU with instrument-specific keywords
    Then: both the common and instrument specific keywords values are available as properties in the
    derived class
    """

    class InstFitsAccess(L0FitsAccess):
        def __init__(self, hdu, name):
            super().__init__(hdu, name)
            self.foo: str = self.header["INST_FOO"]

    fits_obj = InstFitsAccess(hdu_with_complete_common_header, name="foo")
    assert fits_obj.foo == "bar"
    assert fits_obj.elevation == 6.28
    assert fits_obj.azimuth == 3.14
    assert fits_obj.table_angle == 1.23
    assert fits_obj.time_obs == "1988-05-25T01:23:45.678"
    assert fits_obj.name == "foo"
    np.testing.assert_equal(fits_obj.data, np.arange(9).reshape(3, 3))


def test_squeezing_array(extra_axis_hdu_with_complete_common_header):
    """
    Given: an HDU with a 3D array, where one axis has length 1
    When: loading the HDU with the CommonFitsData class
    Then: the data element only contains two axes, the third axis being removed
    """
    fits_obj = L0FitsAccess(extra_axis_hdu_with_complete_common_header)
    assert len(fits_obj.data.shape) == 2


def test_squeezing_array_with_intentional_unitary_axis(
    singleton_axis_hdu_with_complete_common_header,
):
    """
    Given: an HDU with a 2D array, where one axis has length 1
    When: loading the HDU with the CommonFitsData class
    Then: the data element only contains two axes, with the 'squeeze' not taking effect
    """
    fits_obj = L0FitsAccess(singleton_axis_hdu_with_complete_common_header)
    assert len(fits_obj.data.shape) == 2


def test_false_auto_squeeze(three_singleton_axes_hdu):
    """
    Given: an HDU with a 3D array where all axes have length 1
    When: loading the HDU with a FitsAccess class with auto_squeeze set to False
    Then: the full, 3D shape of the data is preserved
    """
    fits_obj = L0FitsAccess(three_singleton_axes_hdu, auto_squeeze=False)
    assert fits_obj.data.shape == (1, 1, 1)


def test_setting_data(hdu_with_complete_common_header):
    """
    Given: a L0FitsAccess object with data and a header
    When: setting the object's data property with a new array
    Then: the object's data and dynamic header keys are correctly updated
    """
    fits_obj = L0FitsAccess(hdu_with_complete_common_header)
    new_array = np.random.random((10, 4))  # Intentionally a different shape
    fits_obj.data = new_array
    np.testing.assert_equal(fits_obj.data, new_array)
    assert fits_obj.header["NAXIS1"] == 4
    assert fits_obj.header["NAXIS2"] == 10



"""
Components of the Constant model
"""
from enum import Enum


class BudName(str, Enum):
    """
    Controlled list of names for constant stems (buds)
    """

    instrument = "INSTRUMENT"
    num_cs_steps = "NUM_CS_STEPS"
    num_modstates = "NUM_MODSTATES"
    proposal_id = "PROPOSAL_ID"
    average_cadence = "AVERAGE_CADENCE"
    maximum_cadence = "MAXIMUM_CADENCE"
    minimum_cadence = "MINIMUM_CADENCE"
    variance_cadence = "VARIANCE_CADENCE"
    time_order = "TIME_ORDER"
    num_dsps_repeats = "NUM_DSPS_REPEATS"
    spectral_line = "SPECTRAL_LINE"



"""
Components of the Tag model.  Stem + Optional Suffix = Tag
"""
from enum import Enum
from typing import Union


class StemName(str, Enum):
    """
    Controlled list of Tag Stems
    """

    output = "OUTPUT"
    input = "INPUT"
    intermediate = "INTERMEDIATE"
    input_dataset = "INPUT_DATASET"
    frame = "FRAME"
    movie = "MOVIE"
    stokes = "STOKES"
    movie_frame = "MOVIE_FRAME"
    task = "TASK"
    cs_step = "CS_STEP"
    modstate = "MODSTATE"
    dsps_repeat = "DSPS_REPEAT"
    calibrated = "CALIBRATED"  # A flag to indicate the data has been calibrated but not yet output


class Tag:
    """
    Controlled methods for creating tags from stems + optional suffixes
    """

    @staticmethod
    def format_tag(stem: Union[StemName, str], *parts):
        if isinstance(stem, Enum):
            stem = stem.value
        parts = [stem, *parts]
        return "_".join([str(part).upper() for part in parts])

    # Static Tags
    @classmethod
    def movie_frame(cls):
        return cls.format_tag(StemName.movie_frame)

    @classmethod
    def input(cls):
        return cls.format_tag(StemName.input)

    @classmethod
    def calibrated(cls) -> str:
        return cls.format_tag(StemName.calibrated)

    @classmethod
    def output(cls):
        return cls.format_tag(StemName.output)

    @classmethod
    def frame(cls):
        return cls.format_tag(StemName.frame)

    @classmethod
    def intermediate(cls):
        return cls.format_tag(StemName.intermediate)

    @classmethod
    def input_dataset(cls):
        return cls.format_tag(StemName.input_dataset)

    @classmethod
    def movie(cls):
        return cls.format_tag(StemName.movie)

    # Dynamic Tags
    @classmethod
    def task(cls, ip_task_type: str):
        return cls.format_tag(StemName.task, ip_task_type)

    @classmethod
    def cs_step(cls, n: int):
        return cls.format_tag(StemName.cs_step, n)

    @classmethod
    def modstate(cls, n: int):
        return cls.format_tag(StemName.modstate, n)

    @classmethod
    def stokes(cls, stokes_state: str) -> str:
        return cls.format_tag(StemName.stokes, stokes_state)

    @classmethod
    def dsps_repeat(cls, dsps_repeat_number: int):
        return cls.format_tag(StemName.dsps_repeat, dsps_repeat_number)



from io import BytesIO

import pytest

from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.tasks.base import WorkflowDataTaskBase


class WorkflowDataTaskBaseTask(WorkflowDataTaskBase):
    def run(self):
        pass


@pytest.fixture(scope="function")
def workflow_data_task(tmp_path, recipe_run_id):
    number_of_files = 10
    tag_string = "WORKFLOW_DATA_TASK"
    tag_object = Tag.input()
    filenames = [f"file_{filenum}.ext" for filenum in range(number_of_files)]
    with WorkflowDataTaskBaseTask(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        task.scratch = WorkflowFileSystem(
            recipe_run_id=recipe_run_id,
            scratch_base_path=tmp_path,
        )
        task.scratch.workflow_base_path = tmp_path / str(recipe_run_id)
        for filename in filenames:
            filepath = task.scratch.workflow_base_path / filename
            filepath.touch()
            task.tag(filepath, tag_string)
            task.tag(filepath, tag_object)

        yield task, number_of_files, filenames, tag_string, tag_object
        task.scratch.purge()
        task.constants.purge()


def test_valid_read_with_strings(workflow_data_task):
    """
    Given: a WorkflowDataTask with tagged data
    When: reading tagged files using a string
    Then: the correct number of files are returned and they have the correct names
    """
    task, number_of_files, filenames, tag_string, _ = workflow_data_task
    task()
    tagged_filepaths = list(task.read(tags=tag_string))
    assert len(tagged_filepaths) == number_of_files
    for tagged_filepath in tagged_filepaths:
        assert tagged_filepath.name in filenames
        assert tagged_filepath.exists()


def test_valid_read_with_tag_object(workflow_data_task):
    """
    Given: a WorkflowDataTask with tagged data
    When: reading tagged files using a string
    Then: the correct number of files are returned and they have the correct names
    """
    task, number_of_files, filenames, _, tag_object = workflow_data_task
    tagged_filepaths = list(task.read(tags=tag_object))
    assert len(tagged_filepaths) == number_of_files
    for tagged_filepath in tagged_filepaths:
        assert tagged_filepath.name in filenames
        assert tagged_filepath.exists()


def test_valid_write_with_bytes(workflow_data_task):
    """
    Given: a WorkflowDataTask
    When: writing a bytes object to disk
    Then: the file is on disk and correctly tagged
    """
    task, _, _, _, _ = workflow_data_task
    relative_path = "bytes_path"
    task.write(file_obj=bytes("abcdefg", "utf-8"), tags="BYTES_OBJECT", relative_path=relative_path)
    assert (task.scratch.workflow_base_path / relative_path).exists()
    assert len(list(task.read(tags="BYTES_OBJECT"))) == 1


def test_valid_write_with_bytesio(workflow_data_task):
    """
    Given: a WorkflowDataTask
    When: writing a BytesIO object to disk
    Then: the file is on disk and correctly tagged
    """
    task, _, _, _, _ = workflow_data_task
    relative_path = "bytesio_path"
    task.write(
        file_obj=BytesIO(bytes("abcdefg", "utf-8")),
        tags="BYTESIO_OBJECT",
        relative_path=relative_path,
    )
    assert (task.scratch.workflow_base_path / relative_path).exists()
    assert len(list(task.read(tags="BYTESIO_OBJECT"))) == 1


def test_write_tags_is_none(workflow_data_task):
    """
    Given: a WorkflowDataTask
    When: writing a file to disk with tags=None
    Then: a ValueError is raised
    """
    task, _, _, _, _ = workflow_data_task
    relative_path = "bytesio_path"
    with pytest.raises(ValueError):
        task.write(file_obj=bytes("abcdefg", "utf-8"), tags=None, relative_path=relative_path)


def test_read_nonexistent_tag(workflow_data_task):
    """
    Given: a WorkflowDataTask
    When: reading from a tag that doesn't exist
    Then: an empty generator is returned
    """
    task, _, _, _, _ = workflow_data_task
    filepaths = task.read(tags="DOES_NOT_EXIST")
    with pytest.raises(StopIteration):
        next(filepaths)


def test_tag_nonexistent_file(workflow_data_task):
    """
    Given: a WorkflowDataTask
    When: trying to tag a file that doesn't exist
    Then: a FileNotFoundError is raised
    """
    task, _, _, _, _ = workflow_data_task
    with pytest.raises(FileNotFoundError):
        task.tag(path=task.scratch.workflow_base_path / "abc.ext", tags="NONEXISTENT_FILE")


def test_tag_not_on_base_path(workflow_data_task):
    """
    Given: a WorkflowDataTask
    When: trying to tag a file that isn't on the workflow base path
    Then: a ValueError is raised
    """
    task, _, _, _, _ = workflow_data_task
    with pytest.raises(ValueError):
        task.tag(path="abc.ext", tags="NOT_ON_BASE_PATH")



"""
Tag cloud manager
"""
import logging
from pathlib import Path
from typing import Iterable
from typing import List
from typing import Set
from typing import Tuple
from typing import Union

from redis import Redis

from dkist_processing_common._util.config import get_mesh_config


class TagDB:
    def __init__(self, recipe_run_id: int, task_name: str, namespace: str = "path"):
        """
        Initialize a connection to a tag database whose instance is uniquely identified by
        the recipe_run_id and whose client connection is identified with the task_name
        """
        self.recipe_run_id = recipe_run_id
        self.task_name = task_name
        host, port = self._db_host_port()
        self._db = self.recipe_run_id % 15
        self.namespace = f"{self.recipe_run_id}:{namespace}:"
        connection_name = f"{self.namespace}{self.task_name}"
        self.db = Redis(db=self._db, host=host, port=port, client_name=connection_name)

    @staticmethod
    def _db_host_port() -> Tuple[str, int]:
        mesh_default = '{"automated-processing-scratch-inventory": {"mesh_address": "localhost", "mesh_port": 6379}}'
        mesh_config = get_mesh_config(default=mesh_default)
        return (
            mesh_config["automated-processing-scratch-inventory"]["mesh_address"],
            mesh_config["automated-processing-scratch-inventory"]["mesh_port"],
        )

    @staticmethod
    def _format_query_result(result: List[bytes]) -> Set[str]:
        return {r.decode("utf8") for r in result}

    def _add_name_space(self, tags: Union[Iterable[str], str]) -> Union[Iterable[str], str]:
        if isinstance(tags, str):
            return self.namespace + tags
        return [f"{self.namespace}{t}" for t in tags]

    def _remove_name_space(self, raw_tags: Union[Iterable[bytes], bytes]) -> Iterable[str]:
        prefix_length = len(self.namespace)
        if isinstance(raw_tags, bytes):
            raw_tags = [raw_tags]
        return [str(k[prefix_length:], "UTF-8") for k in raw_tags]

    def add(self, tag: str, value: str):
        """
        Add values to a tag
        """
        tag = self._add_name_space(tag)
        self.db.sadd(tag, value)

    def clear_tag(self, tag: str) -> None:
        """
        Remove a tag from the database
        """
        tag = self._add_name_space(tag)
        self.db.delete(tag)

    def clear_value(self, value: Union[Path, str]) -> None:
        """
        Remove a value from all tags in the DB
        """
        for tag in self._namespace_keys:
            self.remove(tag, value)

    def remove(self, tag: Union[str, bytes], value: Union[Path, str]):
        """
        Remove a value from a tag
        """
        if isinstance(value, Path):
            value = str(value)
        self.db.srem(tag, value)

    def any(self, tags: Union[Iterable[str], str]) -> Set[str]:
        """
        Return a set of values that match any of the tags
        """
        tags = self._add_name_space(tags)
        r = self.db.sunion(tags)
        if len(r) == 0:
            logging.warning(f"No files found matching ANY tags in {tags=}")
        return self._format_query_result(r)

    def all(self, tags: Union[Iterable[str], str]) -> Set[str]:
        """
        Return a set of values that match all of the tags
        """
        tags = self._add_name_space(tags)
        r = self.db.sinter(tags)
        if len(r) == 0:
            logging.warning(f"No files found matching the set of {tags=}")
        return self._format_query_result(r)

    def tags_for_value(self, value: str) -> Iterable[str]:
        """
        Return list of tags assigned to a value
        """
        value_tags = [t for t in self._namespace_keys if self.db.sismember(t, value)]
        return self._remove_name_space(value_tags)

    def close(self):
        """
        Close the connection to the tag db.  For use at the end of a task
        """
        self.db.close()

    @property
    def _namespace_keys(self) -> List[bytes]:
        return self.db.keys(f"{self.namespace}*")

    @property
    def tags(self) -> Iterable[str]:
        return self._remove_name_space(self._namespace_keys)

    def purge(self) -> None:
        """
        Remove the database of tags.  For use at the end of a workflow.
        """
        if keys := self._namespace_keys:
            self.db.delete(*keys)

    def __repr__(self):
        return f"TagDB(recipe_run_id={self.recipe_run_id}, task_name={self.task_name})"

    def __str__(self):
        return f"{self!r} connected to {self.db}"



"""
Task(s) for the transfer in of data sources for a processing pipeline
"""
import logging
from pathlib import Path
from typing import List

from dkist_processing_common.models.tags import Tag
from dkist_processing_common.tasks.base import WorkflowDataTaskBase
from dkist_processing_common.tasks.mixin.globus import GlobusMixin
from dkist_processing_common.tasks.mixin.globus import GlobusTransferItem
from dkist_processing_common.tasks.mixin.input_dataset import InputDatasetMixin
from dkist_processing_common.tasks.mixin.metadata_store import MetadataStoreMixin

__all__ = ["TransferL0Data"]

logger = logging.getLogger(__name__)


class TransferL0Data(WorkflowDataTaskBase, GlobusMixin, MetadataStoreMixin, InputDatasetMixin):
    """
    Transfers Level 0 data to the scratch store
    """

    def download_input_dataset(self):
        raw_input_dataset = self.metadata_store_input_dataset_document
        self.write(file_obj=raw_input_dataset.encode("utf8"), tags=Tag.input_dataset())

    def format_transfer_items(self) -> List[GlobusTransferItem]:
        transfer_items = []
        bucket = self.input_dataset_bucket
        for frame in self.input_dataset_frames:
            source_path = Path("/", bucket, frame)
            destination_path = self.scratch.absolute_path(frame)
            transfer_items.append(
                GlobusTransferItem(
                    source_path=source_path,
                    destination_path=destination_path,
                    recursive=False,
                )
            )
        return transfer_items

    def tag_input_data(self, transfer_items: List[GlobusTransferItem]):
        scratch_items = [
            self.scratch.scratch_base_path / ti.destination_path for ti in transfer_items
        ]
        for si in scratch_items:
            self.tag(si, tags=[Tag.input(), Tag.frame()])

    def run(self) -> None:
        with self.apm_step("Change Status to InProgress"):
            self.metadata_store_change_recipe_run_to_inprogress()

        with self.apm_step("Download Input Dataset"):
            self.download_input_dataset()

        with self.apm_step("Format Transfer Items"):
            transfer_items = self.format_transfer_items()

        if not transfer_items:
            raise ValueError("No input dataset frames found")

        with self.apm_step("Transfer Inputs via Globus"):
            self.globus_transfer_object_store_to_scratch(
                transfer_items=transfer_items,
                label=f"Transfer Inputs for Recipe Run {self.recipe_run_id}",
            )

        with self.apm_step("Tag Input Data"):
            self.tag_input_data(transfer_items=transfer_items)



"""
Framework for grouping multiple keys and values with arbitrary logic

Defines:
    Stem -> ABC for groupings that depend on both the key and (maybe) value. Subgroups (Petals) are implied but not enforced.

    FlowerPot -> Container for Stem children (Flowers)
"""
from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from collections.abc import Hashable
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List


class FlowerPot:
    def __init__(self):
        self.flowers: List[Stem] = list()

    def __iter__(self):
        return self.flowers.__iter__()

    def __len__(self):
        return self.flowers.__len__()

    def __getitem__(self, item):
        return self.flowers.__getitem__(item)

    def add_dirt(self, key: Hashable, value: Any):
        """
        Send key and value through all Flowers
        """
        if not isinstance(key, Hashable):
            raise TypeError(f"Type of key ({type(key)}) is not hashable")

        for flower in self.flowers:
            flower.update(key, value)


class SpilledDirt:
    """
    A custom class for when a Flower wants the FlowerPot to skip that particular key/value.

    Exists because None, False, [], (), etc. etc. are all valid Flower return values
    """


class Petal:
    def __init__(self, item: tuple):
        self.value = item[0]
        self.keys = item[1]

    def __repr__(self):
        return f"Petal: {{{self.value}: {self.keys}}}"


class Stem(ABC):
    """
    Base group for grouping keys via arbitrary logic on the total collection of keys and values
    """

    def __init__(self, stem_name: Any):
        self.stem_name = stem_name
        self.key_to_petal_dict: Dict[Hashable, Hashable] = dict()

    def update(self, key: Hashable, value: Any):
        """ Ingest a single key/value pair """
        result = self.setter(value)
        if result is not SpilledDirt:
            self.key_to_petal_dict[key] = result

    @property
    def petals(self) -> Iterator[Petal]:
        """ Return subgroups and associated keys """
        petal_to_key_dict = defaultdict(list)
        for key, petal in self.key_to_petal_dict.items():
            petal = self.getter(key)
            petal_to_key_dict[petal].append(key)

        return (Petal(item) for item in petal_to_key_dict.items())

    @property
    def bud(self) -> Petal:
        """ Just the first petal """
        return next(self.petals)

    @abstractmethod
    def setter(self, value: Any) -> Any:
        """ Logic to apply to a single key/value pair on ingest """
        pass

    @abstractmethod
    def getter(self, key: Hashable) -> Hashable:
        """ Logic to apply to all ingested values when picking the Flower """
        pass



from typing import Any
from typing import Hashable

import pytest

from dkist_processing_common.models.flower_pot import FlowerPot
from dkist_processing_common.models.flower_pot import SpilledDirt
from dkist_processing_common.models.flower_pot import Stem


@pytest.fixture()
def simple_flower():
    class Flower(Stem):
        def setter(self, value: Any) -> Any:
            if value < 0:
                return SpilledDirt
            return value % 2

        def getter(self, key: Hashable) -> Hashable:
            return self.key_to_petal_dict[key]

    return Flower(stem_name="simple_flower")


@pytest.fixture()
def simple_flower_pot(simple_flower):
    flower_pot = FlowerPot()
    flower_pot.flowers += [simple_flower]

    return flower_pot


@pytest.fixture()
def simple_key_values():
    return {f"thing{i}": i for i in range(5)}


def test_simple_flower_pot(simple_flower_pot, simple_key_values):
    """
    Given: A FlowerPot with a simple Flower
    When: Updating rock and flower with key: value pairs
    Then: The rock and flower are correctly updated
    """
    assert len(simple_flower_pot) == 1

    flower = simple_flower_pot[0]
    assert flower.stem_name == "simple_flower"

    for k, m in simple_key_values.items():
        simple_flower_pot.add_dirt(k, m)

    petals = sorted(list(flower.petals), key=lambda x: x.value)
    assert len(petals) == 2
    assert petals[0].value == 0
    assert petals[0].keys == ["thing0", "thing2", "thing4"]
    assert petals[1].value == 1
    assert petals[1].keys == ["thing1", "thing3"]


def test_spilled_dirt_flower(simple_flower):
    """
    Given: A simple Flower with logic to deal with SpilledDirt
    When: Updating the flower with a key/value that causes dirt to be spilled
    Then: The Flower ignores the input key/value
    """
    key = "thing0"
    value = -1
    simple_flower.update(key, value)
    assert len(list(simple_flower.petals)) == 0


def test_unhashable_dirt(simple_flower_pot):
    """
    Given: A FlowerPot
    When: Adding dirt with a key that is not hashable
    Then: An Error is raised
    """
    unhashable_key = ["a", "list"]
    value = "never gonna get here"
    with pytest.raises(TypeError):
        simple_flower_pot.add_dirt(unhashable_key, value)



"""
By-frame header keywords that are not instrument specific
"""
from typing import Optional
from typing import Union

from astropy.io import fits

from dkist_processing_common.models.fits_access import FitsAccessBase


class L0FitsAccess(FitsAccessBase):
    def __init__(
        self,
        hdu: Union[fits.ImageHDU, fits.PrimaryHDU, fits.CompImageHDU],
        name: Optional[str] = None,
        auto_squeeze: bool = True,
    ):
        super().__init__(hdu=hdu, name=name, auto_squeeze=auto_squeeze)

        self.elevation: float = self.header["ELEV_ANG"]
        self.azimuth: float = self.header["TAZIMUTH"]
        self.table_angle: float = self.header["TTBLANGL"]
        self.gos_level3_status: str = self.header["LVL3STAT"]
        self.gos_level3_lamp_status: str = self.header["LAMPSTAT"]
        self.scanning_mode: str = self.header["TELSCAN"]
        self.gos_polarizer_status: str = self.header["LVL2STAT"]
        self.gos_polarizer_angle: float = self.header["POLANGLE"]
        self.gos_retarder_status: str = self.header["LVL1STAT"]
        self.gos_retarder_angle: float = self.header["RETANGLE"]
        self.gos_level0_status: str = self.header["LVL0STAT"]
        self.time_obs: str = self.header["DATE-OBS"]
        self.ip_task_type: str = self.header["DKIST004"]
        self.ip_id: str = self.header["ID___004"]
        self.instrument: str = self.header["INSTRUME"]
        self.wavelength: float = self.header["LINEWAV"]
        self.date_begin: str = self.header["DATE-BEG"]
        self.date_end: str = self.header["DATE-END"]
        self.proposal_id: str = self.header["PROP_ID"]
        self.num_dsps_repeats: int = self.header["DKIST008"]
        self.current_dsps_repeat: int = self.header["DKIST009"]



"""
Task(s) for the clean up tasks at the conclusion of a processing pipeline
"""
import logging

from dkist_processing_common.tasks.base import WorkflowDataTaskBase
from dkist_processing_common.tasks.mixin.metadata_store import MetadataStoreMixin


__all__ = ["Teardown"]


logger = logging.getLogger(__name__)


class Teardown(WorkflowDataTaskBase, MetadataStoreMixin):
    """
    Changes the status of the recipe run to "COMPLETEDSUCCESSFULLY"
    Deletes the scratch directory containing all data from this pipeline run
    """

    def run(self) -> None:
        with self.apm_step("Change Recipe Run to Complete Successfully"):
            self.metadata_store_change_recipe_run_to_completed_successfully()

        logger.info(f"Removing data and tags for recipe run {self.recipe_run_id}")

        with self.apm_step("Remove Data and Tags"):
            self.scratch.purge()

        with self.apm_step("Remove Constants"):
            self.constants.purge()



import pytest

from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.tasks.teardown import Teardown


class TeardownTest(Teardown):
    def metadata_store_change_recipe_run_to_completed_successfully(self):
        pass


@pytest.fixture(scope="function")
def teardown_task(tmp_path, recipe_run_id):
    number_of_files = 10
    tag_object = Tag.output()
    filenames = [f"file_{filenum}.ext" for filenum in range(number_of_files)]
    with TeardownTest(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        task.scratch = WorkflowFileSystem(
            recipe_run_id=recipe_run_id,
            scratch_base_path=tmp_path,
        )
        task.scratch.workflow_base_path = tmp_path / str(recipe_run_id)
        for filename in filenames:
            filepath = task.scratch.workflow_base_path / filename
            filepath.touch()
            task.tag(filepath, tag_object)

        yield task, filenames, tag_object


def test_purge_data(teardown_task):
    """
    :Given: a Teardown task with files and tags linked to it
    :When: running the purge_data method
    :Then: all the files are deleted and the tags are removed
    """
    task, filenames, tag_object = teardown_task
    tagged_data = task.read(tags=tag_object)
    for filepath in tagged_data:
        assert filepath.exists()
    task()
    for filepath in tagged_data:
        assert not filepath.exists()
    post_purge_tagged_data = list(task.read(tags=tag_object))
    assert len(post_purge_tagged_data) == 0



"""
Global test fixtures
"""
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from random import randint
from typing import Dict
from typing import List
from typing import Tuple

import pytest
from astropy.io import fits
from dkist_data_simulator.dataset import key_function
from dkist_data_simulator.spec122 import Spec122Dataset
from dkist_header_validator import spec122_validator

from dkist_processing_common._util.constants import Constants
from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common._util.tags import TagDB
from dkist_processing_common.models.graphql import InputDatasetResponse
from dkist_processing_common.models.graphql import RecipeInstanceResponse
from dkist_processing_common.models.graphql import RecipeRunResponse
from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess


@pytest.fixture()
def recipe_run_id():
    return randint(0, 99999)


@pytest.fixture()
def tag_db(recipe_run_id) -> TagDB:
    t = TagDB(recipe_run_id=recipe_run_id, task_name="test_tags")
    yield t
    t.purge()
    t.close()


@pytest.fixture()
def tag_db2(recipe_run_id) -> TagDB:
    """
    Another instance of a tag db in the same redis db
    """
    recipe_run_id = recipe_run_id + 15  # same db number but different namespace
    t = TagDB(recipe_run_id=recipe_run_id, task_name="test_tags2")
    yield t
    t.purge()
    t.close()


@pytest.fixture(params=[None, "use_tmp_path"])
def workflow_file_system(request, recipe_run_id, tmp_path) -> Tuple[WorkflowFileSystem, int, Path]:
    if request.param == "use_tmp_path":
        path = tmp_path
    else:
        path = request.param
    wkflow_fs = WorkflowFileSystem(
        recipe_run_id=recipe_run_id,
        task_name="wkflow_fs_test",
        scratch_base_path=path,
    )
    yield wkflow_fs, recipe_run_id, tmp_path
    wkflow_fs.purge(ignore_errors=True)
    tmp_path.rmdir()
    wkflow_fs.close()


@pytest.fixture()
def constants(recipe_run_id) -> Constants:
    constants = Constants(recipe_run_id=recipe_run_id, task_name="test_constants")
    yield constants
    constants.purge()
    constants.close()


@pytest.fixture()
def complete_common_header():
    """
    A header with some common by-frame keywords and a single instrument specific one
    """
    hdu = fits.PrimaryHDU()
    hdu.header["ELEV_ANG"] = 6.28
    hdu.header["TAZIMUTH"] = 3.14
    hdu.header["TTBLANGL"] = 1.23
    hdu.header["DATE-OBS"] = "1988-05-25T01:23:45.678"
    hdu.header["INST_FOO"] = "bar"
    hdu.header["DKIST004"] = "ip task type"
    hdu.header["ID___004"] = "ip id"
    hdu.header["LVL2STAT"] = "polarizer"
    hdu.header["POLANGLE"] = 31.2
    hdu.header["LVL1STAT"] = "retarder"
    hdu.header["RETANGLE"] = 6.66
    hdu.header["LVL0STAT"] = "dark shutter"
    hdu.header["INSTRUME"] = "VBI"
    hdu.header["LINEWAV"] = 1080.0
    hdu.header["DATE-BEG"] = "2020-03-13T00:00:00.000"
    hdu.header["DATE-END"] = "2021-08-15T00:00:00.000"
    hdu.header["PROP_ID"] = "PROPOSAL_ID1"
    hdu.header["LVL3STAT"] = "Clear"
    hdu.header["LAMPSTAT"] = "On"
    hdu.header["TELSCAN"] = "Raster"
    hdu.header["DKIST008"] = 1
    hdu.header["DKIST009"] = 1

    return hdu.header


class CalibrationSequenceDataset(Spec122Dataset):
    def __init__(
        self,
        array_shape: Tuple[int, ...],
        time_delta: float,
        instrument="visp",
    ):
        self.num_frames_per_CS_step = 5

        # Make up a Calibration sequence. Mostly random except for two clears and two darks at start and end, which
        # we want to test
        self.pol_status = [
            "clear",
            "clear",
            "Polarizer",
            "Polarizer",
            "Polarizer",
            "clear",
            "clear",
        ]
        self.pol_theta = [0.0, 0.0, 60.0, 60.0, 120.0, 0.0, 0.0]
        self.ret_status = ["clear", "clear", "clear", "SiO2", "clear", "clear", "clear"]
        self.ret_theta = [0.0, 0.0, 0.0, 45.0, 0.0, 0.0, 0.0]
        self.dark_status = [
            "DarkShutter",
            "FieldStop (2.8arcmin)",
            "FieldStop (2.8arcmin)",
            "FieldStop (2.8arcmin)",
            "FieldStop (2.8arcmin)",
            "FieldStop (2.8arcmin)",
            "DarkShutter",
        ]

        self.num_steps = len(self.pol_theta)
        dataset_shape = (self.num_steps * self.num_frames_per_CS_step,) + array_shape[1:]
        super().__init__(
            dataset_shape,
            array_shape,
            time_delta,
            instrument=instrument,
            start_time=datetime(2010, 1, 1, 1, 1, 1),
        )
        self.add_constant_key("DKIST004", "polcal")
        self.add_constant_key("TELEVATN", 6.28)
        self.add_constant_key("ID___013", "PROPOSAL_ID1")
        self.add_constant_key("PAC__002", "clear")
        self.add_constant_key("PAC__003", "on")

    @property
    def cs_step(self) -> int:
        return self.index // self.num_frames_per_CS_step

    @key_function("PAC__004")
    def polarizer_status(self, key: str) -> str:
        return self.pol_status[self.cs_step]

    @key_function("PAC__005")
    def polarizer_angle(self, key: str) -> float:
        return self.pol_theta[self.cs_step]

    @key_function("PAC__006")
    def retarder_status(self, key: str) -> str:
        return self.ret_status[self.cs_step]

    @key_function("PAC__007")
    def retarder_angle(self, key: str) -> float:
        return self.ret_theta[self.cs_step]

    @key_function("PAC__008")
    def gos_level3_status(self, key: str) -> str:
        return self.dark_status[self.cs_step]


class NonPolCalDataset(Spec122Dataset):
    def __init__(self):
        super().__init__(
            dataset_shape=(4, 2, 2),
            array_shape=(1, 2, 2),
            time_delta=1,
            instrument="visp",
            start_time=datetime(2010, 1, 1, 1, 1, 1),
        )  # Instrument doesn't matter
        self.add_constant_key("DKIST004", "dark")  # Anything that's not polcal
        self.add_constant_key("ID___013", "PROPOSAL_ID1")
        self.add_constant_key("TELEVATN", 6.28)
        self.add_constant_key("PAC__002", "clear")
        self.add_constant_key("PAC__003", "on")
        self.add_constant_key("PAC__004", "Polarizer")
        self.add_constant_key("PAC__005", 0.0)
        self.add_constant_key("PAC__006", "clear")
        self.add_constant_key("PAC__007", 0.0)
        self.add_constant_key("PAC__008", "DarkShutter")


@pytest.fixture(scope="session")
def grouped_cal_sequence_headers() -> Dict[int, List[L0FitsAccess]]:
    ds = CalibrationSequenceDataset(array_shape=(1, 2, 2), time_delta=2.0)
    header_list = [
        spec122_validator.validate_and_translate_to_214_l0(d.header(), return_type=fits.HDUList)[
            0
        ].header
        for d in ds
    ]
    expected_cs_dict = defaultdict(list)
    for i in range(ds.num_steps):
        for j in range(ds.num_frames_per_CS_step):
            expected_cs_dict[i].append(L0FitsAccess.from_header(header_list.pop(0)))

    return expected_cs_dict


@pytest.fixture(scope="session")
def non_polcal_headers() -> List[L0FitsAccess]:
    ds = NonPolCalDataset()
    header_list = [
        spec122_validator.validate_and_translate_to_214_l0(d.header(), return_type=fits.HDUList)[
            0
        ].header
        for d in ds
    ]
    obj_list = [L0FitsAccess.from_header(h) for h in header_list]
    return obj_list


class FakeGQLClient:
    def __init__(self, *args, **kwargs):
        pass

    @staticmethod
    def execute_gql_query(**kwargs):
        query_base = kwargs["query_base"]

        if query_base == "recipeRuns":
            return [
                RecipeRunResponse(
                    recipeInstanceId=1,
                    recipeInstance=RecipeInstanceResponse(
                        recipeId=1,
                        inputDataset=InputDatasetResponse(
                            inputDatasetId=1,
                            isActive=True,
                            inputDatasetDocument='{"bucket": "bucket-name", "parameters": [{"parameterName": "", "parameterValues": [{"parameterValueId": 1, "parameterValue": "[[1,2,3],[4,5,6],[7,8,9]]", "parameterValueStartDate": "1/1/2000"}]}], "frames": ["objectKey1", "objectKey2", "objectKeyN"]}',
                        ),
                    ),
                )
            ]

    @staticmethod
    def execute_gql_mutation(**kwargs):
        ...



"""
Common tasks and bases
"""
from dkist_processing_common.tasks.base import *
from dkist_processing_common.tasks.l1_output_data import *
from dkist_processing_common.tasks.parse_l0_input_data import *
from dkist_processing_common.tasks.teardown import *
from dkist_processing_common.tasks.transfer_input_data import *



"""
Helper functions for retrieving configurations
"""
import json
from collections import defaultdict
from os import environ
from typing import Union


def get_config(
    key: str, default: Union[None, bool, str, dict, list, int, float] = None, is_json: bool = False
) -> Union[None, bool, str, dict, list, int, float]:
    value = environ.get(key, default)
    if is_json:
        return json.loads(value)
    return value


def get_mesh_config(default: str = "null") -> dict:
    # Environment variable indicating how to connect to dependencies on the service mesh
    MESH_CONFIG = get_config("MESH_CONFIG", default=default, is_json=True)
    host_port_default = {"mesh_address": "localhost", "mesh_port": 5672}
    mesh_default = defaultdict(lambda: host_port_default, {})
    MESH_CONFIG = MESH_CONFIG or mesh_default
    return MESH_CONFIG



class SpectralLine:
    def __init__(self, name: str, wavelength: float, wavemin: float, wavemax: float):
        self.name = name
        self.wavelength = wavelength
        self.wavemin = wavemin
        self.wavemax = wavemax
        if not (wavemin <= wavelength <= wavemax):
            raise ValueError(
                "SpectralLine.wavemin must be smaller than SpectralLine.wavelength, which must be smaller than SpectralLine.wavemax"
            )

    @property
    def linewidth(self) -> float:
        return self.wavemax - self.wavemin

    def __repr__(self):
        return f"SpectralLine: name={self.name}, wavelength={self.wavelength}, wavemin={self.wavemin}, wavemax={self.wavemax}"


def find_associated_line(wavelength: float, lines: [SpectralLine]) -> str:
    """
    Given a list of SpectralLine objects, find the one that contains the wavelength of the frame
    within its range, and return its name.
    """
    matching_lines = []
    for line in lines:
        if line.wavemin <= wavelength <= line.wavemax:
            matching_lines.append(line)
    if len(matching_lines) == 1:
        return matching_lines[0].name
    if len(matching_lines) > 1:
        raise ValueError(
            f"{len(matching_lines)} valid spectral lines were found for this frame: {matching_lines=}, {wavelength=}"
        )
    raise ValueError(f"No valid line found for wavelength {wavelength}")



"""
GraphQL Data models for the metadata store api
"""
from dataclasses import dataclass
from os import environ

AUTH_TOKEN = environ.get("GQL_AUTH_TOKEN")


@dataclass
class RecipeRunMutation:
    recipeRunId: int
    recipeRunStatusId: int
    authToken: str = AUTH_TOKEN


@dataclass
class RecipeRunStatusResponse:
    recipeRunStatusId: int


@dataclass
class RecipeRunStatusQuery:
    recipeRunStatusName: str


@dataclass
class CreateRecipeRunStatusResponse:
    recipeRunStatus: RecipeRunStatusResponse


@dataclass
class RecipeRunStatusMutation:
    recipeRunStatusName: str
    isComplete: bool
    recipeRunStatusDescription: str
    authToken: str = AUTH_TOKEN


@dataclass
class InputDatasetResponse:
    inputDatasetId: int
    isActive: bool
    inputDatasetDocument: str


@dataclass
class RecipeInstanceResponse:
    inputDataset: InputDatasetResponse
    recipeId: int


@dataclass
class RecipeRunResponse:
    recipeInstance: RecipeInstanceResponse
    recipeInstanceId: int


@dataclass
class RecipeRunQuery:
    recipeRunId: int


@dataclass
class DatasetCatalogReceiptAccountMutation:
    """
    Dataclass used to write the dataset_catalog_receipt_account record for the run.
    It sets an expected object count for a dataset so that dataset inventory creation
    doesn't happen until all objects are transferred and inventoried.
    """

    datasetId: str
    expectedObjectCount: int
    authToken: str = AUTH_TOKEN


@dataclass
class RecipeRunProvenanceMutation:
    inputDatasetId: int
    isTaskManual: bool
    recipeRunId: int
    taskName: str
    libraryVersions: str
    workflowVersion: str
    codeVersion: str = None
    authToken: str = AUTH_TOKEN



from pathlib import Path

import pytest

from dkist_processing_common.tasks.l1_output_data import PublishCatalogMessages


@pytest.fixture
def publish_catalog_messages_task(recipe_run_id):
    with PublishCatalogMessages(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        proposal_id = "publish_catalog_messages_proposal_id"
        task.constants["PROPOSAL_ID"] = proposal_id
        yield task, proposal_id
        task.constants.purge()


def test_frame_messages(publish_catalog_messages_task):
    """
    :Given: a PublishCatalogMessages task
    :When: creating frame messages
    :Then: the attributes are correctly populated
    """
    task, proposal_id = publish_catalog_messages_task
    filenames = [f"test_frame_{i}.ext" for i in range(10)]
    print(filenames)
    filepaths = [Path(f"a/b/c/{filename}") for filename in filenames]
    frame_messages = task.frame_messages(paths=filepaths)
    for message in frame_messages:
        assert message.bucket == "data"
        assert proposal_id in message.objectName
        assert message.conversationId == str(task.recipe_run_id)


def test_object_messages(publish_catalog_messages_task):
    """
    :Given: a PublishCatalogMessages task
    :When: creating object messages
    :Then: the attributes are correctly populated
    """
    task, proposal_id = publish_catalog_messages_task
    object_type = "test_type"
    filenames = [f"test_object_{i}.ext" for i in range(10)]
    print(filenames)
    filepaths = [Path(f"a/b/c/{filename}") for filename in filenames]
    object_messages = task.object_messages(paths=filepaths, object_type=object_type)
    for message in object_messages:
        assert message.bucket == "data"
        assert proposal_id in message.objectName
        assert message.conversationId == str(task.recipe_run_id)
        assert message.objectType == object_type
        assert message.groupId == task.dataset_id



import pytest

from dkist_processing_common.models.spectral_line import find_associated_line
from dkist_processing_common.models.spectral_line import SpectralLine

line_a = SpectralLine(name="A", wavelength=10.0, wavemin=9.0, wavemax=11.0)
line_b = SpectralLine(name="B", wavelength=15.0, wavemin=14.5, wavemax=15.5)
line_c = SpectralLine(name="C", wavelength=17.0, wavemin=15.1, wavemax=20.0)


def test_spectral_line():
    """
    :Given: a valid SpectralLine object
    :When: examining its properties
    :Then: the properties have the expected values
    """
    assert line_a.name == "A"
    assert line_a.wavelength == 10.0
    assert line_a.wavemin == 9.0
    assert line_a.wavemax == 11.0
    assert line_a.linewidth == 2


def test_invalid_spectral_line():
    """
    :Given: an invalid SpectralLine object with wavemax < wavemin
    :When: instantiating the object
    :Then: an error is raised
    """
    with pytest.raises(ValueError):
        SpectralLine(name="name", wavelength=15.0, wavemin=14.0, wavemax=13.0)


def test_find_associated_line():
    """
    :Given: a collection of SpectralLines and a wavelength contained within one of them
    :When: searching for the associated line
    :Then: the correct line is found
    """
    associated_line = find_associated_line(wavelength=14.75, lines=[line_a, line_b])
    assert associated_line == "B"


def test_find_associated_line_too_many_matches():
    """
    :Given: a collection of SpectralLines and a wavelength contained within *TWO* of them
    :When: searching for the associated line
    :Then: an error is raised
    """
    with pytest.raises(ValueError):
        find_associated_line(wavelength=15.25, lines=[line_a, line_b, line_c])


def test_find_associated_line_no_matches():
    """
    :Given: a collection of SpectralLines and a wavelength contained within *NONE* of them
    :When: searching for the associated line
    :Then: an error is raised
    """
    with pytest.raises(ValueError):
        find_associated_line(wavelength=21.0, lines=[line_a, line_b, line_c])



import json
from datetime import datetime
from pathlib import Path
from uuid import uuid4

import pytest

from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common.models.tags import Tag
from dkist_processing_common.tasks.base import WorkflowDataTaskBase
from dkist_processing_common.tasks.mixin.input_dataset import InputDatasetMixin


class Task(WorkflowDataTaskBase, InputDatasetMixin):
    def run(self):
        pass


INPUT_DATASET_STRING = '{"bucket": "bucket-name", "parameters": [{"parameterName": "param_name", "parameterValues": [{"parameterValueId": 1, "parameterValue": "[[1,2,3],[4,5,6],[7,8,9]]", "parameterValueStartDate": "2000-01-01"}]}], "frames": ["objectKey1", "objectKey2", "objectKeyN"]}'


@pytest.fixture
def input_dataset(tmp_path, recipe_run_id):
    with Task(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        task.scratch = WorkflowFileSystem(
            recipe_run_id=recipe_run_id,
            scratch_base_path=tmp_path,
        )
        task.scratch.workflow_base_path = tmp_path / str(recipe_run_id)
        file_path = task.scratch.workflow_base_path / Path(f"{uuid4().hex[:6]}.ext")
        file_path.write_text(data=INPUT_DATASET_STRING)
        task.tag(path=file_path, tags=Tag.input_dataset())
        input_dataset_object = json.loads(INPUT_DATASET_STRING)
        yield task, input_dataset_object
        task.scratch.purge()
        task.constants.purge()


@pytest.fixture
def multiple_input_datasets(tmp_path, recipe_run_id):
    with Task(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        task.scratch = WorkflowFileSystem(
            recipe_run_id=recipe_run_id,
            scratch_base_path=tmp_path,
        )
        task.scratch.workflow_base_path = tmp_path / str(recipe_run_id)
        for _ in range(2):
            file_path = task.scratch.workflow_base_path / f"{uuid4().hex[:6]}.ext"
            file_path.write_text(data=INPUT_DATASET_STRING)
            task.tag(path=file_path, tags=Tag.input_dataset())
        yield task
        task.scratch.purge()
        task.constants.purge()


def test_input_dataset_document(input_dataset):
    """
    Given: a task with the InputDatasetMixin
    When: reading the input dataset document
    Then: it matches the string used to create the input dataset
    """
    task, input_dataset_object = input_dataset
    task()
    assert task.input_dataset_document == input_dataset_object


def test_multiple_input_datasets(multiple_input_datasets):
    """
    Given: a task with the InputDatasetMixin and multiple tagged input datasets
    When: reading the input dataset document
    Then: an error is raised
    """
    task = multiple_input_datasets
    with pytest.raises(ValueError):
        task.input_dataset_document


def test_no_input_datasets(input_dataset):
    """
    Given: a task with the InputDatasetMixin
    When: deleting the input dataset tag and then trying to read the input dataset document
    Then: the input dataset document is empty
    """
    task, _ = input_dataset
    task.scratch._tag_db.clear_tag(tag=Tag.input_dataset())
    assert task.input_dataset_document == dict()


def test_input_dataset_frames(input_dataset):
    """
    Given: a task with the InputDatasetMixin
    When: getting the frames in the input dataset
    Then: it matches the frames used to create the input dataset
    """
    task, input_dataset_object = input_dataset
    assert task.input_dataset_frames == input_dataset_object.get("frames")


def test_input_dataset_bucket(input_dataset):
    """
    Given: a task with the InputDatasetMixin
    When: getting the bucket in the input dataset
    Then: it matches the bucket used to create the input dataset
    """
    task, input_dataset_object = input_dataset
    assert task.input_dataset_bucket == input_dataset_object.get("bucket")


def test_input_dataset_parameters(input_dataset):
    """
    Given: a task with the InputDatasetMixin
    When: getting the parameters in the input dataset
    Then: the names of the parameters match the keys in the returned dictionary
    """
    task, input_dataset_object = input_dataset
    for key, value in task.input_dataset_parameters.items():
        assert key == input_dataset_object["parameters"][0]["parameterName"]


def test_input_dataset_parameters_get(input_dataset):
    """
    Given: a task with the InputDatasetMixin
    When: asking for a specific parameter value
    Then: the correct value or default is returned
    """
    task, input_dataset_object = input_dataset
    assert task.input_dataset_parameters_get("param_name") == [[1, 2, 3], [4, 5, 6], [7, 8, 9]]


def test_input_dataset_parameters_get_default(input_dataset):
    """
    Given: a task with the InputDatasetMixin
    When: asking for a specific parameter value that does not exist
    Then: the default value is returned
    """
    task, input_dataset_object = input_dataset
    assert task.input_dataset_parameters_get("not_here", default=13) == 13


def test_input_dataset_parameters_get_out_of_range(input_dataset):
    """
    Given: a task with the InputDatasetMixin
    When: asking for a specific parameter value at a time that is too far in the past
    Then: an error is raised
    """
    task, input_dataset_object = input_dataset
    with pytest.raises(ValueError):
        _ = task.input_dataset_parameters_get("param_name", start_date=datetime(1776, 7, 4))



"""
Mixin for a WorkflowDataTaskBase subclass which implements interservice bus access functionality
"""
import json
from os import environ
from typing import Any
from typing import Dict
from typing import List

from talus import DurableBlockingProducerWrapper

from dkist_processing_common._util.config import get_mesh_config


class InterserviceBusMixin:
    @property
    def _interservice_bus_client_retry_config(self) -> dict:
        retry_configuration = environ.get(
            "RETRY_CONFIG",
            '{"retry_delay":1,"retry_backoff":2,"retry_jitter":[1,10],"retry_max_delay":300}',
        )
        retry_configuration = json.loads(retry_configuration)
        retry_configuration["retry_jitter"] = tuple(retry_configuration["retry_jitter"])
        return retry_configuration

    @property
    def _interservice_bus_client_config(self) -> dict:
        mesh_config = get_mesh_config()
        return {
            "rabbitmq_host": mesh_config["interservice-bus"]["mesh_address"],
            "rabbitmq_port": mesh_config["interservice-bus"]["mesh_port"],
            "rabbitmq_user": environ.get("ISB_USERNAME", "guest"),
            "rabbitmq_pass": environ.get("ISB_PASSWORD", "guest"),
        }

    @staticmethod
    def _interservice_bus_parse_bindings(messages: List[Any]) -> List[Dict[str, str]]:
        # @message_class decorated messages only
        # transform in to a single list
        all_bindings = []
        for m in messages:
            all_bindings += m.binding()
        # dedup
        result = []
        for binding in all_bindings:
            if binding not in result:
                result.append(binding)
        return result

    def interservice_bus_publish(self, messages: List[Any]):
        # @message_class decorated messages only
        bindings = self._interservice_bus_parse_bindings(messages=messages)
        with DurableBlockingProducerWrapper(
            producer_queue_bindings=bindings,
            publish_exchange="master.direct.x",
            **self._interservice_bus_client_config,
            **self._interservice_bus_client_retry_config,
        ) as producer:
            for message in messages:
                producer.publish_message(message)



"""
Pre-made flower that produces tag based on a single header key
"""
from dkist_processing_common.models.flower_pot import Stem
from dkist_processing_common.parsers.l0_fits_access import L0FitsAccess


class SingleValueSingleKeyFlower(Stem):
    """ Flower that just passes through a single header value """

    def __init__(self, tag_stem_name: str, metadata_key: str):
        super().__init__(stem_name=tag_stem_name)
        self.metadata_key = metadata_key

    def setter(self, fits_obj: L0FitsAccess):
        return getattr(fits_obj, self.metadata_key)

    def getter(self, key):
        return self.key_to_petal_dict[key]
